# This is Multi-reference timbre encoder

import torch
from torch import nn
from torch.nn.utils import remove_weight_norm, weight_norm
from module.attentions import MultiHeadAttention


class MRTE(nn.Module):
    def __init__(
        self,
        content_enc_channels=192,
        hidden_size=512,
        out_channels=192,
        kernel_size=5,
        n_heads=4,
        ge_layer=2,
    ):
        super(MRTE, self).__init__()
        self.cross_attention = MultiHeadAttention(hidden_size, hidden_size, n_heads)
        self.c_pre = nn.Conv1d(content_enc_channels, hidden_size, 1)
        self.text_pre = nn.Conv1d(content_enc_channels, hidden_size, 1)
        self.c_post = nn.Conv1d(hidden_size, out_channels, 1)

    def forward(self, ssl_enc, ssl_mask, text, text_mask, ge, test=None):
        if ge == None:
            ge = 0
        attn_mask = text_mask.unsqueeze(2) * ssl_mask.unsqueeze(-1)

        ssl_enc = self.c_pre(ssl_enc * ssl_mask)
        text_enc = self.text_pre(text * text_mask)
        if test != None:
            if test == 0:
                x = (
                    self.cross_attention(
                        ssl_enc * ssl_mask, text_enc * text_mask, attn_mask
                    )
                    + ssl_enc
                    + ge
                )
            elif test == 1:
                x = ssl_enc + ge
            elif test == 2:
                x = (
                    self.cross_attention(
                        ssl_enc * 0 * ssl_mask, text_enc * text_mask, attn_mask
                    )
                    + ge
                )
            else:
                raise ValueError("test should be 0,1,2")
        else:
            x = (
                self.cross_attention(
                    ssl_enc * ssl_mask, text_enc * text_mask, attn_mask
                )
                + ssl_enc
                + ge
            )
        x = self.c_post(x * ssl_mask)
        return x


class SpeakerEncoder(torch.nn.Module):
    def __init__(
        self,
        mel_n_channels=80,
        model_num_layers=2,
        model_hidden_size=256,
        model_embedding_size=256,
    ):
        super(SpeakerEncoder, self).__init__()
        self.lstm = nn.LSTM(
            mel_n_channels, model_hidden_size, model_num_layers, batch_first=True
        )
        self.linear = nn.Linear(model_hidden_size, model_embedding_size)
        self.relu = nn.ReLU()

    def forward(self, mels):
        self.lstm.flatten_parameters()
        _, (hidden, _) = self.lstm(mels.transpose(-1, -2))
        embeds_raw = self.relu(self.linear(hidden[-1]))
        return embeds_raw / torch.norm(embeds_raw, dim=1, keepdim=True)


class MELEncoder(nn.Module):
    def __init__(
        self,
        in_channels,
        out_channels,
        hidden_channels,
        kernel_size,
        dilation_rate,
        n_layers,
    ):
        super().__init__()
        self.in_channels = in_channels
        self.out_channels = out_channels
        self.hidden_channels = hidden_channels
        self.kernel_size = kernel_size
        self.dilation_rate = dilation_rate
        self.n_layers = n_layers

        self.pre = nn.Conv1d(in_channels, hidden_channels, 1)
        self.enc = WN(hidden_channels, kernel_size, dilation_rate, n_layers)
        self.proj = nn.Conv1d(hidden_channels, out_channels, 1)

    def forward(self, x):
        # print(x.shape,x_lengths.shape)
        x = self.pre(x)
        x = self.enc(x)
        x = self.proj(x)
        return x


class WN(torch.nn.Module):
    def __init__(self, hidden_channels, kernel_size, dilation_rate, n_layers):
        super(WN, self).__init__()
        assert kernel_size % 2 == 1
        self.hidden_channels = hidden_channels
        self.kernel_size = kernel_size
        self.dilation_rate = dilation_rate
        self.n_layers = n_layers

        self.in_layers = torch.nn.ModuleList()
        self.res_skip_layers = torch.nn.ModuleList()

        for i in range(n_layers):
            dilation = dilation_rate**i
            padding = int((kernel_size * dilation - dilation) / 2)
            in_layer = nn.Conv1d(
                hidden_channels,
                2 * hidden_channels,
                kernel_size,
                dilation=dilation,
                padding=padding,
            )
            in_layer = weight_norm(in_layer)
            self.in_layers.append(in_layer)

            # last one is not necessary
            if i < n_layers - 1:
                res_skip_channels = 2 * hidden_channels
            else:
                res_skip_channels = hidden_channels

            res_skip_layer = torch.nn.Conv1d(hidden_channels, res_skip_channels, 1)
            res_skip_layer = weight_norm(res_skip_layer, name="weight")
            self.res_skip_layers.append(res_skip_layer)

    def forward(self, x):
        output = torch.zeros_like(x)
        n_channels_tensor = torch.IntTensor([self.hidden_channels])

        for i in range(self.n_layers):
            x_in = self.in_layers[i](x)

            acts = fused_add_tanh_sigmoid_multiply(x_in, n_channels_tensor)

            res_skip_acts = self.res_skip_layers[i](acts)
            if i < self.n_layers - 1:
                res_acts = res_skip_acts[:, : self.hidden_channels, :]
                x = x + res_acts
                output = output + res_skip_acts[:, self.hidden_channels :, :]
            else:
                output = output + res_skip_acts
        return output

    def remove_weight_norm(self):
        for l in self.in_layers:
            remove_weight_norm(l)
        for l in self.res_skip_layers:
            remove_weight_norm(l)


@torch.jit.script
def fused_add_tanh_sigmoid_multiply(input, n_channels):
    n_channels_int = n_channels[0]
    t_act = torch.tanh(input[:, :n_channels_int, :])
    s_act = torch.sigmoid(input[:, n_channels_int:, :])
    acts = t_act * s_act
    return acts


if __name__ == "__main__":
    content_enc = torch.randn(3, 192, 100)
    content_mask = torch.ones(3, 1, 100)
    ref_mel = torch.randn(3, 128, 30)
    ref_mask = torch.ones(3, 1, 30)
    model = MRTE()
    out = model(content_enc, content_mask, ref_mel, ref_mask)
    print(out.shape)



import re
from typing import List, Tuple, Optional, Union
from gallama.data_classes.data_class import TextTag, ArtifactTag


class StreamParser:
    def __init__(self):
        self.buffer = ""
        self.current_element = None
        self.current_tag = None
        self.current_content = ""
        self.MALFORMED_CHECK_LENGTH = 300
        self.in_answer_tag = False
        # self.xml_prefix = "```xml\n"
        self.xml_prefix = ""
        self.root_key = "<answer>"
        self.full_root_key = f"{self.xml_prefix}{self.root_key}"
        self.tag_pattern = re.compile(r'(<artifact\s+(?:(?:identifier|type|title|language)="[^"]*"\s*)*>)|(<text>)')
        self.comment_pattern = re.compile(r'<!--.*?-->', re.DOTALL)

    def process_stream(self, new_data: str) -> List[Tuple[Union[TextTag, ArtifactTag], str]]:
        self.buffer += new_data
        self.remove_comments()

        results = []

        while True:
            if not self.in_answer_tag:
                start_index = self.buffer.find(self.full_root_key)
                if start_index != -1:
                    self.buffer = self.buffer[start_index + len(self.full_root_key):]
                    self.in_answer_tag = True
                elif len(self.buffer) >= self.MALFORMED_CHECK_LENGTH:
                    results.append((TextTag(), self.buffer))
                    self.buffer = ""
                else:
                    break

            if self.current_element is None:
                match = self.tag_pattern.search(self.buffer)
                if match:
                    element_type = "artifact" if match.group(1) else "text"
                    start_index = match.start()
                    tag_length = match.end() - match.start()

                    self.current_element = element_type
                    if element_type == "artifact":
                        self.current_tag = self._parse_artifact_tag(self.buffer[start_index:match.end()])
                    else:
                        self.current_tag = TextTag()

                    self.buffer = self.buffer[start_index + tag_length:]
                    self.current_content = ""
                elif len(self.buffer) >= self.MALFORMED_CHECK_LENGTH:
                    results.append((TextTag(), self.buffer))
                    self.buffer = ""
                else:
                    break
            elif len(self.buffer) <= len(f'</{self.current_element}>'):
                break
            else:
                end_tag = f'</{self.current_element}>'
                end_index = self.buffer.find(end_tag)
                if end_index != -1:
                    content = self.buffer[:end_index]
                    results.append((self.current_tag, content))
                    self.buffer = self.buffer[end_index + len(end_tag):]
                    self.current_element = None
                    self.current_tag = None
                    self.current_content = ""
                else:
                    if self.buffer:
                        content_chunk = self.buffer[:len(self.buffer) - len(end_tag)]
                        self.buffer = self.buffer[len(self.buffer) - len(end_tag):]
                        results.append((self.current_tag, content_chunk))
                    break

        return results

    def remove_comments(self):
        comment_pattern = re.compile(r'<!--.*?-->', re.DOTALL)
        self.buffer = comment_pattern.sub('', self.buffer)

    def _parse_artifact_tag(self, tag_content: str) -> ArtifactTag:
        attributes = re.findall(r'(\w+)="([^"]*)"', tag_content)
        attr_dict = dict(attributes)
        return ArtifactTag(
            artifact_type=attr_dict.get('type', 'code'),
            identifier=attr_dict.get('identifier', ''),
            title=attr_dict.get('title', ''),
            language=attr_dict.get('language')
        )

    def get_current_state(self) -> Tuple[Optional[str], str]:
        return self.current_element, self.current_content

    def parse_full_response(self, response: str) -> List[Tuple[Union[TextTag, ArtifactTag], str]]:
        if not response.startswith(self.xml_prefix):
            response = f"{self.xml_prefix}{response}"
        response = self.comment_pattern.sub('', response)
        return self.process_stream(response)


# modified from https://github.com/yangdongchao/SoundStorm/blob/master/soundstorm/s1/AR/models/t2s_model.py
# reference: https://github.com/lifeiteng/vall-e
import torch
from tqdm import tqdm

from AR.modules.embedding_onnx import SinePositionalEmbedding
from AR.modules.embedding_onnx import TokenEmbedding
from AR.modules.transformer_onnx import LayerNorm
from AR.modules.transformer_onnx import TransformerEncoder
from AR.modules.transformer_onnx import TransformerEncoderLayer
from torch import nn
from torch.nn import functional as F
from torchmetrics.classification import MulticlassAccuracy

default_config = {
    "embedding_dim": 512,
    "hidden_dim": 512,
    "num_head": 8,
    "num_layers": 12,
    "num_codebook": 8,
    "p_dropout": 0.0,
    "vocab_size": 1024 + 1,
    "phoneme_vocab_size": 512,
    "EOS": 1024,
}

inf_tensor_value = torch.FloatTensor([-float("Inf")]).float()

def logits_to_probs(
    logits,
    previous_tokens = None,
    temperature: float = 1.0,
    top_k = None,
    top_p = None,
    repetition_penalty: float = 1.0,
):
    previous_tokens = previous_tokens.squeeze()
    if previous_tokens is not None and repetition_penalty != 1.0:
        previous_tokens = previous_tokens.long()
        score = torch.gather(logits, dim=0, index=previous_tokens)
        score = torch.where(
            score < 0, score * repetition_penalty, score / repetition_penalty
        )
        logits.scatter_(dim=0, index=previous_tokens, src=score)

    if top_p is not None and top_p < 1.0:
        sorted_logits, sorted_indices = torch.sort(logits, descending=True)
        cum_probs = torch.cumsum(
            torch.nn.functional.softmax(sorted_logits, dim=-1), dim=-1
        )
        sorted_indices_to_remove = cum_probs > top_p
        sorted_indices_to_remove[0] = False  # keep at least one option
        indices_to_remove = sorted_indices_to_remove.scatter(
            dim=0, index=sorted_indices, src=sorted_indices_to_remove
        )
        logits = logits.masked_fill(indices_to_remove, -float("Inf"))

    logits = logits / max(temperature, 1e-5)

    if top_k is not None:
        v, _ = torch.topk(logits, top_k)
        pivot = v.select(-1, -1).unsqueeze(-1)
        logits = torch.where(logits < pivot, inf_tensor_value, logits)

    probs = torch.nn.functional.softmax(logits, dim=-1)
    return probs


def multinomial_sample_one_no_sync(
    probs_sort
):  # Does multinomial sampling without a cuda synchronization
    q = torch.randn_like(probs_sort)
    return torch.argmax(probs_sort / q, dim=-1, keepdim=True).to(dtype=torch.int)


def sample(
    logits,
    previous_tokens,
    **sampling_kwargs,
):
    probs = logits_to_probs(
        logits=logits, previous_tokens=previous_tokens, **sampling_kwargs
    )
    idx_next = multinomial_sample_one_no_sync(probs)
    return idx_next, probs


class OnnxEncoder(nn.Module):
    def __init__(self, ar_text_embedding, bert_proj, ar_text_position):
        super().__init__()
        self.ar_text_embedding = ar_text_embedding
        self.bert_proj = bert_proj
        self.ar_text_position = ar_text_position
    
    def forward(self, x, bert_feature):
        x = self.ar_text_embedding(x)
        x = x + self.bert_proj(bert_feature.transpose(1, 2))
        return self.ar_text_position(x)


class T2SFirstStageDecoder(nn.Module):
    def __init__(self, ar_audio_embedding, ar_audio_position, h, ar_predict_layer, loss_fct, ar_accuracy_metric,
    top_k, early_stop_num, num_layers):
        super().__init__()
        self.ar_audio_embedding = ar_audio_embedding
        self.ar_audio_position = ar_audio_position
        self.h = h
        self.ar_predict_layer = ar_predict_layer
        self.loss_fct = loss_fct
        self.ar_accuracy_metric = ar_accuracy_metric
        self.top_k = top_k
        self.early_stop_num = early_stop_num
        self.num_layers = num_layers
    
    def forward(self, x, prompt):
        y = prompt
        x_example = x[:,:,0] * 0.0
        #N, 1, 512
        cache = {
            "all_stage": self.num_layers,
            "k": None,
            "v": None,
            "y_emb": None,
            "first_infer": 1,
            "stage": 0,
        }

        y_emb = self.ar_audio_embedding(y)

        cache["y_emb"] = y_emb
        y_pos = self.ar_audio_position(y_emb)

        xy_pos = torch.concat([x, y_pos], dim=1)

        y_example = y_pos[:,:,0] * 0.0
        x_attn_mask = torch.matmul(x_example.transpose(0, 1) , x_example).bool()
        y_attn_mask = torch.ones_like(torch.matmul(y_example.transpose(0, 1), y_example), dtype=torch.int64)
        y_attn_mask = torch.cumsum(y_attn_mask, dim=1) - torch.cumsum(
            torch.ones_like(y_example.transpose(0, 1), dtype=torch.int64), dim=0
        )
        y_attn_mask = y_attn_mask > 0

        x_y_pad = torch.matmul(x_example.transpose(0, 1), y_example).bool()
        y_x_pad = torch.matmul(y_example.transpose(0, 1), x_example).bool()
        x_attn_mask_pad = torch.cat([x_attn_mask, torch.ones_like(x_y_pad)], dim=1)
        y_attn_mask = torch.cat([y_x_pad, y_attn_mask], dim=1)
        xy_attn_mask = torch.concat([x_attn_mask_pad, y_attn_mask], dim=0)
        cache["k"] = torch.matmul(x_attn_mask_pad[0].float().unsqueeze(-1), torch.zeros((1, 512)))\
        .unsqueeze(1).repeat(self.num_layers, 1, 1, 1)
        cache["v"] = torch.matmul(x_attn_mask_pad[0].float().unsqueeze(-1), torch.zeros((1, 512)))\
        .unsqueeze(1).repeat(self.num_layers, 1, 1, 1)

        xy_dec = self.h(xy_pos, mask=xy_attn_mask, cache=cache)
        logits = self.ar_predict_layer(xy_dec[:, -1])
        samples = sample(logits[0], y, top_k=self.top_k, top_p=1.0, repetition_penalty=1.35)[0].unsqueeze(0)

        y = torch.concat([y, samples], dim=1)

        return y, cache["k"], cache["v"], cache["y_emb"], x_example


class T2SStageDecoder(nn.Module):
    def __init__(self, ar_audio_embedding, ar_audio_position, h, ar_predict_layer, loss_fct, ar_accuracy_metric,
    top_k, early_stop_num, num_layers):
        super().__init__()
        self.ar_audio_embedding = ar_audio_embedding
        self.ar_audio_position = ar_audio_position
        self.h = h
        self.ar_predict_layer = ar_predict_layer
        self.loss_fct = loss_fct
        self.ar_accuracy_metric = ar_accuracy_metric
        self.top_k = top_k
        self.early_stop_num = early_stop_num
        self.num_layers = num_layers

    def forward(self, y, k, v, y_emb, x_example):
        cache = {
            "all_stage": self.num_layers,
            "k": torch.nn.functional.pad(k, (0, 0, 0, 0, 0, 1)),
            "v": torch.nn.functional.pad(v, (0, 0, 0, 0, 0, 1)),
            "y_emb": y_emb,
            "first_infer": 0,
            "stage": 0,
        }

        y_emb = torch.cat(
            [cache["y_emb"], self.ar_audio_embedding(y[:, -1:])], 1
        )
        cache["y_emb"] = y_emb
        y_pos = self.ar_audio_position(y_emb)

        xy_pos = y_pos[:, -1:]
        
        y_example = y_pos[:,:,0] * 0.0

        xy_attn_mask = torch.cat([x_example, y_example], dim=1)
        xy_attn_mask = torch.zeros_like(xy_attn_mask, dtype=torch.bool)

        xy_dec = self.h(xy_pos, mask=xy_attn_mask, cache=cache)
        logits = self.ar_predict_layer(xy_dec[:, -1])
        samples = sample(logits[0], y, top_k=self.top_k, top_p=1.0, repetition_penalty=1.35)[0].unsqueeze(0)

        y = torch.concat([y, samples], dim=1)

        return y, cache["k"], cache["v"], cache["y_emb"], logits, samples


class Text2SemanticDecoder(nn.Module):
    def __init__(self, config, norm_first=False, top_k=3):
        super(Text2SemanticDecoder, self).__init__()
        self.model_dim = config["model"]["hidden_dim"]
        self.embedding_dim = config["model"]["embedding_dim"]
        self.num_head = config["model"]["head"]
        self.num_layers = config["model"]["n_layer"]
        self.norm_first = norm_first
        self.vocab_size = config["model"]["vocab_size"]
        self.phoneme_vocab_size = config["model"]["phoneme_vocab_size"]
        self.p_dropout = float(config["model"]["dropout"])
        self.EOS = config["model"]["EOS"]
        self.norm_first = norm_first
        assert self.EOS == self.vocab_size - 1
        self.bert_proj = nn.Linear(1024, self.embedding_dim)
        self.ar_text_embedding = TokenEmbedding(self.embedding_dim, self.phoneme_vocab_size, self.p_dropout)
        self.ar_text_position = SinePositionalEmbedding(self.embedding_dim, dropout=0.1, scale=False, alpha=True)
        self.ar_audio_embedding = TokenEmbedding(self.embedding_dim, self.vocab_size, self.p_dropout)
        self.ar_audio_position = SinePositionalEmbedding(self.embedding_dim, dropout=0.1, scale=False, alpha=True)
        self.h = TransformerEncoder(
            TransformerEncoderLayer(
                d_model=self.model_dim,
                nhead=self.num_head,
                dim_feedforward=self.model_dim * 4,
                dropout=0.1,
                batch_first=True,
                norm_first=norm_first,
            ),
            num_layers=self.num_layers,
            norm=LayerNorm(self.model_dim) if norm_first else None,
        )
        self.ar_predict_layer = nn.Linear(self.model_dim, self.vocab_size, bias=False)
        self.loss_fct = nn.CrossEntropyLoss(reduction="sum")
        self.ar_accuracy_metric = MulticlassAccuracy(
            self.vocab_size,
            top_k=top_k,
            average="micro",
            multidim_average="global",
            ignore_index=self.EOS,
        )
        self.top_k = torch.LongTensor([1])
        self.early_stop_num = torch.LongTensor([-1])

    def init_onnx(self):
        self.onnx_encoder = OnnxEncoder(self.ar_text_embedding, self.bert_proj, self.ar_text_position)
        self.first_stage_decoder = T2SFirstStageDecoder(self.ar_audio_embedding, self.ar_audio_position, self.h, 
            self.ar_predict_layer, self.loss_fct, self.ar_accuracy_metric, self.top_k, self.early_stop_num,
            self.num_layers)
        self.stage_decoder = T2SStageDecoder(self.ar_audio_embedding, self.ar_audio_position, self.h, 
            self.ar_predict_layer, self.loss_fct, self.ar_accuracy_metric, self.top_k, self.early_stop_num,
            self.num_layers)

    def forward(self, x, prompts, bert_feature):
        early_stop_num = self.early_stop_num
        prefix_len = prompts.shape[1]

        x = self.onnx_encoder(x, bert_feature)
        y, k, v, y_emb, stage, x_example = self.first_stage_decoder(x, prompts)

        stop = False
        for idx in range(1, 1500):
            enco = self.stage_decoder(y, k, v, y_emb, stage, x_example)
            y, k, v, y_emb, stage, logits, samples = enco
            if early_stop_num != -1 and (y.shape[1] - prefix_len) > early_stop_num:
                stop = True
            if torch.argmax(logits, dim=-1)[0] == self.EOS or samples[0, 0] == self.EOS:
                stop = True
            if stop:
                break
        y[0, -1] = 0
        return y, idx

    def infer(self, x, prompts, bert_feature):
        top_k = self.top_k
        early_stop_num = self.early_stop_num

        x = self.onnx_encoder(x, bert_feature)

        y = prompts
        prefix_len = y.shape[1]
        x_len = x.shape[1]
        x_example = x[:,:,0] * 0.0
        x_attn_mask = torch.matmul(x_example.transpose(0, 1), x_example)
        x_attn_mask = torch.zeros_like(x_attn_mask, dtype=torch.bool)

        stop = False
        cache = {
            "all_stage": self.num_layers,
            "k": [None] * self.num_layers,
            "v": [None] * self.num_layers,
            "y_emb": None,
            "first_infer": 1,
            "stage": 0,
        }
        for idx in range(1500):
            if cache["first_infer"] == 1:
                y_emb = self.ar_audio_embedding(y)
            else:
                y_emb = torch.cat(
                    [cache["y_emb"], self.ar_audio_embedding(y[:, -1:])], 1
                )
            cache["y_emb"] = y_emb
            y_pos = self.ar_audio_position(y_emb)
            if cache["first_infer"] == 1:
                xy_pos = torch.concat([x, y_pos], dim=1)
            else:
                xy_pos = y_pos[:, -1:]
            y_len = y_pos.shape[1]
            if cache["first_infer"] == 1:
                x_attn_mask_pad = F.pad(x_attn_mask, (0, y_len), value=True)
                y_attn_mask = F.pad(
                    torch.triu(torch.ones(y_len, y_len, dtype=torch.bool), diagonal=1),
                    (x_len, 0), value=False
                )
                xy_attn_mask = torch.concat([x_attn_mask_pad, y_attn_mask], dim=0)
            else:
                xy_attn_mask = torch.zeros((1, x_len + y_len), dtype=torch.bool)
            xy_dec = self.h(xy_pos, mask=xy_attn_mask, cache=cache)
            logits = self.ar_predict_layer(xy_dec[:, -1])
            samples = sample(logits[0], y, top_k=top_k, top_p=1.0, repetition_penalty=1.35)[0].unsqueeze(0)
            if early_stop_num != -1 and (y.shape[1] - prefix_len) > early_stop_num:
                stop = True
            if torch.argmax(logits, dim=-1)[0] == self.EOS or samples[0, 0] == self.EOS:
                stop = True
            if stop:
                if prompts.shape[1] == y.shape[1]:
                    y = torch.concat([y, torch.zeros_like(samples)], dim=1)
                break
            y = torch.concat([y, samples], dim=1)
            cache["first_infer"] = 0
        return y, idx


# Copyright (c) 2021 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import re

from .num import verbalize_digit

# 规范化固话/手机号码
# 手机
# http://www.jihaoba.com/news/show/13680
# 移动：139、138、137、136、135、134、159、158、157、150、151、152、188、187、182、183、184、178、198
# 联通：130、131、132、156、155、186、185、176
# 电信：133、153、189、180、181、177
RE_MOBILE_PHONE = re.compile(
    r"(?<!\d)((\+?86 ?)?1([38]\d|5[0-35-9]|7[678]|9[89])\d{8})(?!\d)")
RE_TELEPHONE = re.compile(
    r"(?<!\d)((0(10|2[1-3]|[3-9]\d{2})-?)?[1-9]\d{6,7})(?!\d)")

# 全国统一的号码400开头
RE_NATIONAL_UNIFORM_NUMBER = re.compile(r"(400)(-)?\d{3}(-)?\d{4}")


def phone2str(phone_string: str, mobile=True) -> str:
    if mobile:
        sp_parts = phone_string.strip('+').split()
        result = '，'.join(
            [verbalize_digit(part, alt_one=True) for part in sp_parts])
        return result
    else:
        sil_parts = phone_string.split('-')
        result = '，'.join(
            [verbalize_digit(part, alt_one=True) for part in sil_parts])
        return result


def replace_phone(match) -> str:
    """
    Args:
        match (re.Match)
    Returns:
        str
    """
    return phone2str(match.group(0), mobile=False)


def replace_mobile(match) -> str:
    """
    Args:
        match (re.Match)
    Returns:
        str
    """
    return phone2str(match.group(0))



import json
import yaml
from typing import List, Dict, Union, Literal
from gallama.data_classes.data_class import BaseMessage, ChatMLQuery, ToolCall, MultiModalTextContent, MultiModalImageContent
from pydantic import BaseModel
from copy import deepcopy
from gallama.utils.utils import parse_xml_to_dict
from fastapi import HTTPException
from pathlib import Path
from textwrap import dedent
from gallama.data import ARTIFACT_SYSTEM_PROMPT
import uuid

class PromptEngine:
    def __init__(self, prompt_format: str):
        self.system_msg_enabled = False
        self.tool_enabled = False
        self.model_prompt_all = self.get_prompt_token()
        if not self.model_prompt_all.get(prompt_format):
            raise ValueError(f'Prompt format {prompt_format} not found in data/model_token.yaml')

        self.model_prompt = self.model_prompt_all.get(prompt_format)
        self.system_msg_enabled = self.model_prompt.get("system_msg_enabled")
        self.tool_enabled = self.model_prompt.get("tool_enabled")
        self.eos_token_list = self.model_prompt.get("eos_token_list")

    @staticmethod
    def get_prompt_token() -> Dict:
        """Get the absolute path to the data directory."""
        yaml_file = Path(__file__).parent.parent.parent / 'data' / 'model_token.yaml'
        with open(yaml_file, 'r') as file:
            yaml_data = yaml.safe_load(file)
        return yaml_data

    def get_conversation_start_token(self):
        return self.model_prompt.get("conversation_start", "")

    def get_conversation_end_token(self):
        return self.model_prompt.get("conversation_end", "")

    def leading_prompt_token(self):
        return self.model_prompt.get("leading_prompt_token", "")

    def get_user_start_token(self):
        return self.model_prompt.get("user_start", "")

    def get_user_end_token(self):
        return self.model_prompt.get("user_end", "")

    def get_sys_start_token(self):
        return self.model_prompt.get("system_start", "")

    def get_sys_end_token(self):
        return self.model_prompt.get("system_end", "")

    def get_assistant_start_token(self):
        return self.model_prompt.get("assistant_start", "")

    def get_assistant_end_token(self):
        return self.model_prompt.get("assistant_end", "")

    def get_tool_start_token(self):
        return self.model_prompt.get("tool_start", "")

    def get_tool_end_token(self):
        return self.model_prompt.get("tool_end", "")

    def get_tool_result_start_token(self):
        return self.model_prompt.get("tool_result_start", "")

    def get_tool_result_end_token(self):
        return self.model_prompt.get("tool_result_end", "")

    def get_tool_call_start_token(self):
        return self.model_prompt.get("tool_call_start", "")

    def get_tool_call_end_token(self):
        return self.model_prompt.get("tool_call_start", "")

    def get_vision_start_token(self):
        return self.model_prompt.get("vision_start", "")

    def get_vision_end_token(self):
        return self.model_prompt.get("vision_end", "")

    def get_image_pad_token(self):
        return self.model_prompt.get("image_pad", "")

    def _get_role_token(self, role, token_type: Literal["start", "end"]):
        if token_type == "start":
            if role == "system":
                return self.get_sys_start_token()
            elif role == "user":
                return self.get_user_start_token()
            elif role == "assistant":
                return self.get_assistant_start_token()
            elif role == "tool":
                return self.get_tool_start_token()
            elif role == "tool_result":
                return self.get_tool_result_start_token()
            elif role == "tool_call":
                return self.get_tool_call_start_token()
        elif token_type == "end":
            if role == "system":
                return self.get_sys_end_token()
            elif role == "user":
                return self.get_user_end_token()
            elif role == "assistant":
                return self.get_assistant_end_token()
            elif role == "tool":
                return self.get_tool_end_token()
            elif role == "tool_result":
                return self.get_tool_result_end_token()
            elif role == "tool_call":
                return self.get_tool_call_end_token()
        else:
            return ""

    @staticmethod
    def _get_message_type(msg: BaseMessage) -> str:
        if msg.tool_call_id:
            return "tool_result"
        elif msg.role == "tool":
            return "tool_call"
        elif msg.role == "assistant":
            return "assistant"
        elif msg.role == "system":
            return "system"
        elif msg.role == "user":
            return "user"
        else:
            raise ValueError(f"Unknown message type {msg.role}")

    def _format_tool_call(self, tool_call: ToolCall) -> str:
        """
            get string representation of tool call like normal python function calling
            Example Output: get_current_weather(location='Boston', unit='Fahrenheit')
        """

        func_name = tool_call.function.name
        args_str = tool_call.function.arguments.replace('"', '')
        return f"{func_name}({args_str})"

    def _format_tool_result(self, msg: BaseMessage) -> str:
        content = msg.content if msg.content else ""

        try:
            content = self._get_role_token(role="user",token_type="start") + f"Result of tool call reference id {msg.tool_call_id}:\n" + str(json.dumps(json.loads(content), indent=2)) + "\n---\n\n"
        except:
            content = self._get_role_token(role="user",token_type="start") + f"Result of tool call reference id {msg.tool_call_id}:\n" + str(content) + "\n---\n\n"

        return content + self._get_role_token(role="user",token_type="end")

    def _format_tool_result_msg(self, msg: BaseMessage) -> str:
        """one msg might have multiple tool calls"""
        # tool_call_str = f"Please help to call these tool:\n"
        tool_call_str = "---"

        for tool_call in msg.tool_calls:
            tool_call_str += f"Request for tool call with reference id {tool_call.id}:\n"
            tool_call_str += self._format_tool_call(tool_call)
            tool_call_str += "\n\n"

        tool_call_str += "---"
        return tool_call_str

    def _format_tool_msg(self, pydantic_tool_list: Dict[str, BaseModel], pretty:bool = False) -> str:
        # append tools calling if it is a prompt
        tools_json = ("\nBelow are the functions available to you to use. Only use the tool if it is necessary\n")

        if self.tool_enabled:
            tools_json += self.get_tool_start_token()

        for tool_name, tool in pydantic_tool_list.items():
            if pretty:
                tools_json = tools_json + tool_name + ":\n" + str(json.dumps(tool.schema(), indent=2)) + "\n---\n"
            else:
                tools_json = tools_json + tool_name + ":\n" + str(json.dumps(tool.schema())) + "\n---\n"

        if self.tool_enabled:
            tools_json += self.get_tool_end_token()

        return tools_json

    def _format_tool_msg_as_code(self, pydantic_tool_as_code: str) -> str:
        # append tools calling if it is a prompt
        tools_as_code = ("\nBelow are the functions available to you to use.\n"
                      "The function definition and arguments are presented as pydantic class.\n\n")

        if self.tool_enabled:
            tools_as_code += pydantic_tool_as_code + "\n"

        return tools_as_code

    def _get_one_msg(self, msg: BaseMessage) -> str:
        content = msg.content if msg.content else ""

        prompt = ""
        tool_call_str = ""

        if self._get_message_type(msg) == "tool_result":
            return self._format_tool_result(msg)
        elif self._get_message_type(msg) == "tool_call":
            return self._format_tool_result_msg(msg)
        elif self._get_message_type(msg) in ["system", "user"]:
            return content
        elif self._get_message_type(msg) in ["assistant"]:
            # check if there is tool call
            if msg.tool_calls:
                content += self._format_tool_result_msg(msg)

            return content

    def _get_one_msg_grp(self, grp: List[BaseMessage]) -> str:
        prompt = self._get_role_token(role=self._get_message_type(grp[0]), token_type="start")

        for msg in grp:
            prompt = prompt + self._get_one_msg(msg) + "\n\n"

        prompt += self._get_role_token(role=self._get_message_type(grp[0]), token_type="end")
        return prompt

    def _regroup_msg(self, msg_List: List[BaseMessage]):
        # group msg with same role into one array
        regrouped_msg = []
        temp_array = []

        # first create a list of msg with the role converted
        for idx, msg in enumerate(msg_List):
            temp_msg = deepcopy(msg)
            if temp_msg.role == "tool" and not self.tool_enabled:
                temp_msg.role = "user"      # convert to user as this LLM model does not have tool role
            elif temp_msg.role == "system" and not self.system_msg_enabled:
                temp_msg.role = "user"  # convert to user as this LLM model does not have system role

            temp_array.append(temp_msg)

        # now grouping message that was from the same role together
        # this is cause model might have been trained with alternated role prompting
        # and repeat of the same role will not give good response
        previous_msg = msg_List[0]
        grouped_array = []
        for idx, msg in enumerate(temp_array):
            if msg.role == previous_msg.role:
                grouped_array.append(msg)
            else:
                if grouped_array:
                    # create new array
                    regrouped_msg.append(grouped_array.copy())
                    grouped_array = [msg]     # reset temp array
                else:
                    grouped_array.append(msg)

            previous_msg = msg

        # add the last group
        if temp_array:
            regrouped_msg.append(grouped_array.copy())

        # logger.debug("overall regroup:\n" + str(regrouped_msg))
        return regrouped_msg

    def convert_multimodal_content_list_to_string(
        self,
        content: List[Union[MultiModalTextContent, MultiModalImageContent]],
        exllama_vision_token: bool = False,     # to use placeholder token for exllama for not. TODO - refractor
    ) -> str:
        """
        convert multimodal content list to string
        e.g.
            "content": [
                {
                    "type": "image",
                    "image": "https://qianwen-res.oss-cn-beijing.aliyuncs.com/Qwen-VL/assets/demo.jpeg",
                },
                {
                    "type": "text",
                    "text": "Describe this image."
                },
            ]
        to:
            <|vision_start|><|image_pad|><|vision_end|>Describe this image.
        """

        content_str = ""

        for chunk in content:
            if isinstance(chunk, MultiModalTextContent):
                content_str += chunk.text
            elif isinstance(chunk, MultiModalImageContent):
                if not exllama_vision_token:
                    content_str += self.get_vision_start_token() + self.get_image_pad_token() + self.get_vision_end_token()   # TODO
                else:
                    # use a standard token as place holder, TODO - refractor
                    # content_str += "{{IMG-" + f"{uuid.uuid4().hex}" + "}}"
                    content_str += "{{IMG-PlaceHolderTokenHere}}"   #TODO use a constant instead
            else:
                raise ValueError("Unexpected content type ")

        return content_str


    def get_prompt(
        self,
        query: ChatMLQuery,
        pydantic_tool_dict: List[BaseModel] = None,
        answer_format_schema: bool = True,  # whether to add the instruction for tool calling answer schema
        prefix_prompt: str = "",
        leading_prompt: str = "",
        use_thinking: bool = True,
        thinking_template: str = None,
        thinking_response: str = None,
        backend: Literal["exllama", "llama_cpp", "transformers", "embedding"] = "exllama",    # skip model pseudo token and use exllama placeholder token # TODO - code refractoring
        pydantic_tool_code: str = None,     # the code representation of tool
    ) -> str:
        exllama_vision_token = (backend=="exllama")     # vision token in exllama is handled and assigned from the embedding itself

        def _create_leading_prompt(original_prompt, query_leading_prompt, model_leading_prompt):
            # add leading prompt from user or model
            # this is usually the role e.g. assistant:, Tom to Jerry:
            _leading_prompt = ""
            if query_leading_prompt:
                _leading_prompt = model_leading_prompt + query_leading_prompt
            elif model_leading_prompt:
                _leading_prompt = model_leading_prompt

            if not _leading_prompt.endswith("\n"):
                _leading_prompt += "\n"

            return _leading_prompt

        thinking_example = ""

        # thinking_example = dedent("""
        # ### Thinking example:
        # Question: Is dog or cat faster? Answer in Capital letter
        # Apply thinking template:
        # <format_restriction>Any specific format requirement from user</format_restriction>
        # My thought process using the thinking template:
        # <format_restriction>User request for answer in capital letter</format_restriction>
        # Final answer:
        # CAT
        # End of Thinking Example.
        # """).strip()

        msg_list_copy = []

        prompt = self.get_conversation_start_token()     # use arrange to story prompt

        # add the default system prompt for artifact
        if query.artifact != "No":
            prompt += ARTIFACT_SYSTEM_PROMPT

        # standardize multimodal message
        for message in query.messages:
            if isinstance(message.content, list):
                msg_copy = deepcopy(message)
                msg_copy.content = self.convert_multimodal_content_list_to_string(
                    content=msg_copy.content,
                    exllama_vision_token=exllama_vision_token,
                )
                msg_list_copy.append(deepcopy(msg_copy))
            else:
                msg_list_copy.append(deepcopy(message))

        # group message together for the same role
        msg_groups = self._regroup_msg(msg_list_copy)

        # concat all msg
        for idx, msg_group in enumerate(msg_groups):
            prompt = prompt + self._get_one_msg_grp(msg_group)

        # if there is tool, return the prompt with tool
        if pydantic_tool_dict and query.tool_choice != "none":
            if pydantic_tool_code:
                if query.tool_schema_position=="postfix":
                    prompt = prompt + self._format_tool_msg_as_code(pydantic_tool_code)
                else:
                    prompt = self._format_tool_msg_as_code(pydantic_tool_code) + prompt
            else:
                if query.tool_schema_position == "postfix":
                    prompt = prompt + self._format_tool_msg(pydantic_tool_dict)
                else:
                    prompt = self._format_tool_msg(pydantic_tool_dict) + prompt

#             if answer_format_schema:
#                 prompt += """
# IMPORTANT: If you use tool/ functions, please answer using the following schema using json format.
# Each item in the array under "functions_calling" is the "name" of the function to call and its "arguments".
#
# {
#   "functions_calling": [
#     {
#       "name": the name of function/ tool to call,
#       "arguments": {argument_name: argument_value}
#     }
# }
#
#
# If no functions_calling/ tool needed, the response can be:
# {
#   "functions_calling": []
# }
# End of Example of answer with Tool/ Function_calling usage.
#
# """
        # TODO move the thinking_template check to request receiver
        # initialize thinking_template_dict
        thinking_template_dict = {}
        root_key = None
        if thinking_template:
            try:
                thinking_template_dict = parse_xml_to_dict(thinking_template)
                if len(list(thinking_template_dict.keys())) > 1:
                    raise HTTPException(status_code=400, detail=f"thinking_template must be a XML with only 1 root key")

            except Exception as e:
                raise HTTPException(status_code=400, detail=f"thinking_template is not valid XML string")

            # find the XML root key
            root_key = list(thinking_template_dict.keys())[0]

        # prompt creation with thinking_template
        if use_thinking and thinking_template and not thinking_response:
            prompt += _create_leading_prompt(prompt, query.leading_prompt, self.leading_prompt_token())

            prompt += "\nNow, before answering the question, I am required to apply XML thinking template to guide my internal thinking process.\n" + \
                      f"{thinking_example}\n" + \
                      "The thinking template i need to apply to answer this question is as follow:\n" + \
                      thinking_template + "\n" + \
                      f"My thinking using the above XML template as follow:\n\n```xml\n"
            # add ending token
            # prompt += self.get_conversation_end_token()

        elif use_thinking and thinking_template and thinking_response:
            # the thinking result is ready here
            # root key need to be added as we use it for leading prompt in the prompt creation above so it is not part of the response
            prompt += "\nNow, before answering the question, I am required to apply XML thinking template to guide my internal thinking process.\n" + \
                      f"{thinking_example}\n" + \
                      "Now, the thinking template i need to apply to answer this question is:\n" + \
                      thinking_template + "\n" + \
                      f"My thinking using the XML template as follow:\n```xml\n{thinking_response}\n" + \
                      "Now answer the question. Remember that the thinking above is INVISIBLE to user, " + \
                      "and you are PROHIBITED to mentioned to user about the existence of Thinking section. You are ALLOWED to reiterate points from thinking section to user if necessary."

            # add ending token
            prompt += self.get_conversation_end_token()

            # for thinking response, we put it before the leading prompt
            prompt += _create_leading_prompt(prompt, query.leading_prompt, self.leading_prompt_token())

        else:
            # add ending token
            prompt += self.get_conversation_end_token()

            prompt += _create_leading_prompt(prompt, query.leading_prompt, self.leading_prompt_token())


        # add leading_prompt from code which is usually tool related
        # e.g. ```json
        if leading_prompt:      # leading prompt that passed to this function take highest priority
            prompt += leading_prompt + "\n"

        return prefix_prompt + prompt

        # match tool call result #TODO




from fastapi import WebSocket, APIRouter, BackgroundTasks
from ..data_classes import VideoFrame
from typing import Dict
import struct
import asyncio
from ..dependencies import get_video_collection, get_session_config, set_record_start_time, set_clear_history_flag
from ..logger import logger
from pydantic import BaseModel
import os
import time

try:
    from livekit import rtc, api
    from livekit.agents.utils.images import encode, EncodeOptions, ResizeOptions
except ModuleNotFoundError:
    rtc = None
    api = None
    encode = None
    EncodeOptions = None
    ResizeOptions = None

video_frames = get_video_collection()

router = APIRouter(prefix="", tags=["video"])

@router.websocket("/video")
async def websocket_video(websocket: WebSocket):
    await websocket.accept()
    websocket._ping_interval = 60  # Send a ping every 60 seconds
    websocket._ping_timeout = 30   # Wait 30 seconds for a pong response
    try:
        while True:
            data = await websocket.receive_bytes()
            timestamp = None
            logger.info(f"Received video frame")

            session_config = await get_session_config()

            # Check if the message includes a timestamp (first 8 bytes)
            if len(data) >= 8:
                try:
                    # Unpack the first 8 bytes as a double (timestamp)
                    timestamp = struct.unpack('>d', data[:8])[0]
                    frame_data = data[8:]  # The rest is the frame data
                except struct.error:
                    # If unpacking fails, treat the entire message as frame data
                    frame_data = data
            else:
                # If the message is less than 8 bytes, treat it as frame data only
                frame_data = data

            # Create a VideoFrame object
            frame = VideoFrame(frame_data, timestamp)
            video_frames.add_frame(frame, video_max_resolution=session_config.video.video_max_resolution)

            # Optional: Send an acknowledgment back to the client
            await websocket.send_text(f"Frame received with timestamp: {frame.timestamp}")
    except Exception as e:
        logger.info(f"WebSocket closed with exception: {e}")
    finally:
        # Check if the WebSocket is still open before closing it
        try:
            # Attempt to close the WebSocket connection gracefully
            await websocket.close()
        except RuntimeError as e:
            # Ignore the error if the WebSocket is already closed
            logger.error(f"WebSocket already closed: {e}")


class LivekitToken(BaseModel):
    livekit_url: str
    token: str
    save_frame: bool = False

# Global dictionary to store active tasks
active_tasks: Dict[str, asyncio.Task] = {}

@router.post("/v1/video_via_livekit")
async def video_via_livekit(
    token_info: LivekitToken, background_tasks: BackgroundTasks
):
    """Subscribe to the LiveKit room directly and extract one frame per second."""
    livekit_url = token_info.livekit_url
    token = token_info.token
    save_frame = token_info.save_frame

    # Cancel any existing task for this endpoint
    task_key = "video_via_livekit_task"
    if task_key in active_tasks:
        active_tasks[task_key].cancel()
        try:
            await active_tasks[task_key]  # Wait for the task to be cancelled
        except asyncio.CancelledError:
            pass  # Task was successfully cancelled
        del active_tasks[task_key]

    # Start a new background task
    new_task = asyncio.create_task(
        process_video_stream(livekit_url, token, save_frame)
    )
    active_tasks[task_key] = new_task
    background_tasks.add_task(new_task)

    return {"message": "Video processing started"}

async def process_video_stream(livekit_url: str, token: str, save_frame: bool = False):
    """Background task to process the video stream."""
    room = rtc.Room()
    try:
        await room.connect(livekit_url, token)
        session_config = await get_session_config()

        if save_frame:
            save_directory = "/home/remichu/work/ML/gallama/experiment/video_frames"
            os.makedirs(save_directory, exist_ok=True)

        frames_per_second = 1
        frame_interval = 1.0 / frames_per_second
        last_frame_time = 0.0

        async def process_track(track: rtc.RemoteVideoTrack):
            nonlocal last_frame_time
            video_stream = rtc.VideoStream(track)

            async for event in video_stream:
                try:
                    timestamp_us = event.timestamp_us
                    current_time = timestamp_us / 1e6

                    if current_time - last_frame_time >= frame_interval:
                        frame = event.frame
                        image_bytes = encode(
                            frame,
                            EncodeOptions(
                                format="PNG",
                                resize_options=ResizeOptions(
                                    width=480,
                                    height=640,
                                    strategy="scale_aspect_fit",
                                ),
                            ),
                        )

                        if save_frame:
                            image_filename = f"{save_directory}/frame_{timestamp_us}.png"
                            with open(image_filename, "wb") as image_file:
                                image_file.write(image_bytes)

                        frame_obj = VideoFrame(
                            image_bytes,
                            timestamp=time.time()   # use time.time() cause timestamp from livekit is relative timing
                        )
                        video_frames.add_frame(
                            frame_obj,
                            video_max_resolution=session_config.video.video_max_resolution
                        )
                        logger.info(f"Processed frame {current_time}")
                        last_frame_time = current_time
                except Exception as e:
                    logger.error(f"Error processing frame: {e}")

        @room.on("track_subscribed")
        def on_track_subscribed(track: rtc.Track, publication: rtc.RemoteTrackPublication, participant: rtc.RemoteParticipant):

            if track.kind == rtc.TrackKind.KIND_VIDEO:
                asyncio.create_task(process_track(track))

        # @room.on("participant_connected")
        # def on_participant_connected(participant: rtc.RemoteParticipant):
        #     logger.info(f"Participant connected: {participant.identity}")

        # Handle existing participants
        for participant in room.remote_participants.values():
            logger.info(f"Found existing participant: {participant.identity}")
            for publication in participant.track_publications.values():
                if (publication.track and
                    (publication.source == rtc.TrackSource.SOURCE_CAMERA
                    or publication.source == rtc.TrackSource.SOURCE_SCREENSHARE)
                ):
                    logger.info(f"Track subscribed: {publication.sid}")
                    asyncio.create_task(process_track(publication.track))

        while True:
            await asyncio.sleep(0.5)

    except asyncio.CancelledError:
        logger.info("Video processing task cancelled")
        await room.disconnect()
    except Exception as e:
        logger.error(f"Video processing failed: {e}")
        await room.disconnect()
        raise


class RecordTiming(BaseModel):
    start_time: float

@router.post("/v1/record_start_time")
async def record_start_time(
    new_time: RecordTiming,
):
    """ set the start recording time"""
    logger.info(f"set recording start time {new_time.start_time}")
    set_record_start_time(new_time.start_time)

@router.post("/v1/clear_history")
async def clear_history():
    """ set the start recording time"""
    logger.info(f"clear conversation history")
    set_clear_history_flag(True)


# -*- coding: utf-8 -*-

import sys,os
inp_text=                           os.environ.get("inp_text")
inp_wav_dir=                        os.environ.get("inp_wav_dir")
exp_name=                           os.environ.get("exp_name")
i_part=                             os.environ.get("i_part")
all_parts=                          os.environ.get("all_parts")
if "_CUDA_VISIBLE_DEVICES" in os.environ:
     os.environ["CUDA_VISIBLE_DEVICES"] = os.environ["_CUDA_VISIBLE_DEVICES"]
from feature_extractor import cnhubert
opt_dir=                            os.environ.get("opt_dir")
cnhubert.cnhubert_base_path=                os.environ.get("cnhubert_base_dir")
import torch
is_half = eval(os.environ.get("is_half", "True")) and torch.cuda.is_available()

import pdb,traceback,numpy as np,logging
from scipy.io import wavfile
import librosa
now_dir = os.getcwd()
sys.path.append(now_dir)
from tools.my_utils import load_audio,clean_path

# from config import cnhubert_base_path
# cnhubert.cnhubert_base_path=cnhubert_base_path
# inp_text=sys.argv[1]
# inp_wav_dir=sys.argv[2]
# exp_name=sys.argv[3]
# i_part=sys.argv[4]
# all_parts=sys.argv[5]
# os.environ["CUDA_VISIBLE_DEVICES"]=sys.argv[6]
# cnhubert.cnhubert_base_path=sys.argv[7]
# opt_dir="/data/docker/liujing04/gpt-vits/fine_tune_dataset/%s"%exp_name

from time import time as ttime
import shutil
def my_save(fea,path):#####fix issue: torch.save doesn't support chinese path
    dir=os.path.dirname(path)
    name=os.path.basename(path)
    # tmp_path="%s/%s%s.pth"%(dir,ttime(),i_part)
    tmp_path="%s%s.pth"%(ttime(),i_part)
    torch.save(fea,tmp_path)
    shutil.move(tmp_path,"%s/%s"%(dir,name))

hubert_dir="%s/4-cnhubert"%(opt_dir)
wav32dir="%s/5-wav32k"%(opt_dir)
os.makedirs(opt_dir,exist_ok=True)
os.makedirs(hubert_dir,exist_ok=True)
os.makedirs(wav32dir,exist_ok=True)

maxx=0.95
alpha=0.5
if torch.cuda.is_available():
    device = "cuda:0"
# elif torch.backends.mps.is_available():
#     device = "mps"
else:
    device = "cpu"
model=cnhubert.get_model()
# is_half=False
if(is_half==True):
    model=model.half().to(device)
else:
    model = model.to(device)

nan_fails=[]
def name2go(wav_name,wav_path):
    hubert_path="%s/%s.pt"%(hubert_dir,wav_name)
    if(os.path.exists(hubert_path)):return
    tmp_audio = load_audio(wav_path, 32000)
    tmp_max = np.abs(tmp_audio).max()
    if tmp_max > 2.2:
        print("%s-filtered,%s" % (wav_name, tmp_max))
        return
    tmp_audio32 = (tmp_audio / tmp_max * (maxx * alpha*32768)) + ((1 - alpha)*32768) * tmp_audio
    tmp_audio32b = (tmp_audio / tmp_max * (maxx * alpha*1145.14)) + ((1 - alpha)*1145.14) * tmp_audio
    tmp_audio = librosa.resample(
        tmp_audio32b, orig_sr=32000, target_sr=16000
    )#不是重采样问题
    tensor_wav16 = torch.from_numpy(tmp_audio)
    if (is_half == True):
        tensor_wav16=tensor_wav16.half().to(device)
    else:
        tensor_wav16 = tensor_wav16.to(device)
    ssl=model.model(tensor_wav16.unsqueeze(0))["last_hidden_state"].transpose(1,2).cpu()#torch.Size([1, 768, 215])
    if np.isnan(ssl.detach().numpy()).sum()!= 0:
        nan_fails.append((wav_name,wav_path))
        print("nan filtered:%s"%wav_name)
        return
    wavfile.write(
        "%s/%s"%(wav32dir,wav_name),
        32000,
        tmp_audio32.astype("int16"),
    )
    my_save(ssl,hubert_path)

with open(inp_text,"r",encoding="utf8")as f:
    lines=f.read().strip("\n").split("\n")

for line in lines[int(i_part)::int(all_parts)]:
    try:
        # wav_name,text=line.split("\t")
        wav_name, spk_name, language, text = line.split("|")
        wav_name=clean_path(wav_name)
        if (inp_wav_dir != "" and inp_wav_dir != None):
            wav_name = os.path.basename(wav_name)
            wav_path = "%s/%s"%(inp_wav_dir, wav_name)

        else:
            wav_path=wav_name
            wav_name = os.path.basename(wav_name)
        name2go(wav_name,wav_path)
    except:
        print(line,traceback.format_exc())

if(len(nan_fails)>0 and is_half==True):
    is_half=False
    model=model.float()
    for wav in nan_fails:
        try:
            name2go(wav[0],wav[1])
        except:
            print(wav_name,traceback.format_exc())



# Copyright (c) 2021 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import List
from typing import Tuple

import jieba_fast as jieba
from pypinyin import lazy_pinyin
from pypinyin import Style


class ToneSandhi:
    def __init__(self):
        self.must_neural_tone_words = {
            "麻烦",
            "麻利",
            "鸳鸯",
            "高粱",
            "骨头",
            "骆驼",
            "马虎",
            "首饰",
            "馒头",
            "馄饨",
            "风筝",
            "难为",
            "队伍",
            "阔气",
            "闺女",
            "门道",
            "锄头",
            "铺盖",
            "铃铛",
            "铁匠",
            "钥匙",
            "里脊",
            "里头",
            "部分",
            "那么",
            "道士",
            "造化",
            "迷糊",
            "连累",
            "这么",
            "这个",
            "运气",
            "过去",
            "软和",
            "转悠",
            "踏实",
            "跳蚤",
            "跟头",
            "趔趄",
            "财主",
            "豆腐",
            "讲究",
            "记性",
            "记号",
            "认识",
            "规矩",
            "见识",
            "裁缝",
            "补丁",
            "衣裳",
            "衣服",
            "衙门",
            "街坊",
            "行李",
            "行当",
            "蛤蟆",
            "蘑菇",
            "薄荷",
            "葫芦",
            "葡萄",
            "萝卜",
            "荸荠",
            "苗条",
            "苗头",
            "苍蝇",
            "芝麻",
            "舒服",
            "舒坦",
            "舌头",
            "自在",
            "膏药",
            "脾气",
            "脑袋",
            "脊梁",
            "能耐",
            "胳膊",
            "胭脂",
            "胡萝",
            "胡琴",
            "胡同",
            "聪明",
            "耽误",
            "耽搁",
            "耷拉",
            "耳朵",
            "老爷",
            "老实",
            "老婆",
            "老头",
            "老太",
            "翻腾",
            "罗嗦",
            "罐头",
            "编辑",
            "结实",
            "红火",
            "累赘",
            "糨糊",
            "糊涂",
            "精神",
            "粮食",
            "簸箕",
            "篱笆",
            "算计",
            "算盘",
            "答应",
            "笤帚",
            "笑语",
            "笑话",
            "窟窿",
            "窝囊",
            "窗户",
            "稳当",
            "稀罕",
            "称呼",
            "秧歌",
            "秀气",
            "秀才",
            "福气",
            "祖宗",
            "砚台",
            "码头",
            "石榴",
            "石头",
            "石匠",
            "知识",
            "眼睛",
            "眯缝",
            "眨巴",
            "眉毛",
            "相声",
            "盘算",
            "白净",
            "痢疾",
            "痛快",
            "疟疾",
            "疙瘩",
            "疏忽",
            "畜生",
            "生意",
            "甘蔗",
            "琵琶",
            "琢磨",
            "琉璃",
            "玻璃",
            "玫瑰",
            "玄乎",
            "狐狸",
            "状元",
            "特务",
            "牲口",
            "牙碜",
            "牌楼",
            "爽快",
            "爱人",
            "热闹",
            "烧饼",
            "烟筒",
            "烂糊",
            "点心",
            "炊帚",
            "灯笼",
            "火候",
            "漂亮",
            "滑溜",
            "溜达",
            "温和",
            "清楚",
            "消息",
            "浪头",
            "活泼",
            "比方",
            "正经",
            "欺负",
            "模糊",
            "槟榔",
            "棺材",
            "棒槌",
            "棉花",
            "核桃",
            "栅栏",
            "柴火",
            "架势",
            "枕头",
            "枇杷",
            "机灵",
            "本事",
            "木头",
            "木匠",
            "朋友",
            "月饼",
            "月亮",
            "暖和",
            "明白",
            "时候",
            "新鲜",
            "故事",
            "收拾",
            "收成",
            "提防",
            "挖苦",
            "挑剔",
            "指甲",
            "指头",
            "拾掇",
            "拳头",
            "拨弄",
            "招牌",
            "招呼",
            "抬举",
            "护士",
            "折腾",
            "扫帚",
            "打量",
            "打算",
            "打点",
            "打扮",
            "打听",
            "打发",
            "扎实",
            "扁担",
            "戒指",
            "懒得",
            "意识",
            "意思",
            "情形",
            "悟性",
            "怪物",
            "思量",
            "怎么",
            "念头",
            "念叨",
            "快活",
            "忙活",
            "志气",
            "心思",
            "得罪",
            "张罗",
            "弟兄",
            "开通",
            "应酬",
            "庄稼",
            "干事",
            "帮手",
            "帐篷",
            "希罕",
            "师父",
            "师傅",
            "巴结",
            "巴掌",
            "差事",
            "工夫",
            "岁数",
            "屁股",
            "尾巴",
            "少爷",
            "小气",
            "小伙",
            "将就",
            "对头",
            "对付",
            "寡妇",
            "家伙",
            "客气",
            "实在",
            "官司",
            "学问",
            "学生",
            "字号",
            "嫁妆",
            "媳妇",
            "媒人",
            "婆家",
            "娘家",
            "委屈",
            "姑娘",
            "姐夫",
            "妯娌",
            "妥当",
            "妖精",
            "奴才",
            "女婿",
            "头发",
            "太阳",
            "大爷",
            "大方",
            "大意",
            "大夫",
            "多少",
            "多么",
            "外甥",
            "壮实",
            "地道",
            "地方",
            "在乎",
            "困难",
            "嘴巴",
            "嘱咐",
            "嘟囔",
            "嘀咕",
            "喜欢",
            "喇嘛",
            "喇叭",
            "商量",
            "唾沫",
            "哑巴",
            "哈欠",
            "哆嗦",
            "咳嗽",
            "和尚",
            "告诉",
            "告示",
            "含糊",
            "吓唬",
            "后头",
            "名字",
            "名堂",
            "合同",
            "吆喝",
            "叫唤",
            "口袋",
            "厚道",
            "厉害",
            "千斤",
            "包袱",
            "包涵",
            "匀称",
            "勤快",
            "动静",
            "动弹",
            "功夫",
            "力气",
            "前头",
            "刺猬",
            "刺激",
            "别扭",
            "利落",
            "利索",
            "利害",
            "分析",
            "出息",
            "凑合",
            "凉快",
            "冷战",
            "冤枉",
            "冒失",
            "养活",
            "关系",
            "先生",
            "兄弟",
            "便宜",
            "使唤",
            "佩服",
            "作坊",
            "体面",
            "位置",
            "似的",
            "伙计",
            "休息",
            "什么",
            "人家",
            "亲戚",
            "亲家",
            "交情",
            "云彩",
            "事情",
            "买卖",
            "主意",
            "丫头",
            "丧气",
            "两口",
            "东西",
            "东家",
            "世故",
            "不由",
            "不在",
            "下水",
            "下巴",
            "上头",
            "上司",
            "丈夫",
            "丈人",
            "一辈",
            "那个",
            "菩萨",
            "父亲",
            "母亲",
            "咕噜",
            "邋遢",
            "费用",
            "冤家",
            "甜头",
            "介绍",
            "荒唐",
            "大人",
            "泥鳅",
            "幸福",
            "熟悉",
            "计划",
            "扑腾",
            "蜡烛",
            "姥爷",
            "照顾",
            "喉咙",
            "吉他",
            "弄堂",
            "蚂蚱",
            "凤凰",
            "拖沓",
            "寒碜",
            "糟蹋",
            "倒腾",
            "报复",
            "逻辑",
            "盘缠",
            "喽啰",
            "牢骚",
            "咖喱",
            "扫把",
            "惦记",
        }
        self.must_not_neural_tone_words = {
            "男子",
            "女子",
            "分子",
            "原子",
            "量子",
            "莲子",
            "石子",
            "瓜子",
            "电子",
            "人人",
            "虎虎",
            "幺幺",
            "干嘛",
            "学子",
            "哈哈",
            "数数",
            "袅袅",
            "局地",
            "以下",
            "娃哈哈",
            "花花草草",
            "留得",
            "耕地",
            "想想",
            "熙熙",
            "攘攘",
            "卵子",
            "死死",
            "冉冉",
            "恳恳",
            "佼佼",
            "吵吵",
            "打打",
            "考考",
            "整整",
            "莘莘",
            "落地",
            "算子",
            "家家户户",
            "青青",
        }
        self.punc = "：，；。？！“”‘’':,;.?!"

    # the meaning of jieba pos tag: https://blog.csdn.net/weixin_44174352/article/details/113731041
    # e.g.
    # word: "家里"
    # pos: "s"
    # finals: ['ia1', 'i3']
    def _neural_sandhi(self, word: str, pos: str, finals: List[str]) -> List[str]:
        # reduplication words for n. and v. e.g. 奶奶, 试试, 旺旺
        for j, item in enumerate(word):
            if (
                j - 1 >= 0
                and item == word[j - 1]
                and pos[0] in {"n", "v", "a"}
                and word not in self.must_not_neural_tone_words
            ):
                finals[j] = finals[j][:-1] + "5"
        ge_idx = word.find("个")
        if len(word) >= 1 and word[-1] in "吧呢哈啊呐噻嘛吖嗨呐哦哒额滴哩哟喽啰耶喔诶":
            finals[-1] = finals[-1][:-1] + "5"
        elif len(word) >= 1 and word[-1] in "的地得":
            finals[-1] = finals[-1][:-1] + "5"
        # e.g. 走了, 看着, 去过
        elif len(word) == 1 and word in "了着过" and pos in {"ul", "uz", "ug"}:
            finals[-1] = finals[-1][:-1] + "5"
        elif (
            len(word) > 1
            and word[-1] in "们子"
            and pos in {"r", "n"}
            and word not in self.must_not_neural_tone_words
        ):
            finals[-1] = finals[-1][:-1] + "5"
        # e.g. 桌上, 地下, 家里
        elif len(word) > 1 and word[-1] in "上下里" and pos in {"s", "l", "f"}:
            finals[-1] = finals[-1][:-1] + "5"
        # e.g. 上来, 下去
        elif len(word) > 1 and word[-1] in "来去" and word[-2] in "上下进出回过起开":
            finals[-1] = finals[-1][:-1] + "5"
        # 个做量词
        elif (
            ge_idx >= 1
            and (word[ge_idx - 1].isnumeric() or word[ge_idx - 1] in "几有两半多各整每做是")
        ) or word == "个":
            finals[ge_idx] = finals[ge_idx][:-1] + "5"
        else:
            if (
                word in self.must_neural_tone_words
                or word[-2:] in self.must_neural_tone_words
            ):
                finals[-1] = finals[-1][:-1] + "5"

        word_list = self._split_word(word)
        finals_list = [finals[: len(word_list[0])], finals[len(word_list[0]) :]]
        for i, word in enumerate(word_list):
            # conventional neural in Chinese
            if (
                word in self.must_neural_tone_words
                or word[-2:] in self.must_neural_tone_words
            ):
                finals_list[i][-1] = finals_list[i][-1][:-1] + "5"
        finals = sum(finals_list, [])
        return finals

    def _bu_sandhi(self, word: str, finals: List[str]) -> List[str]:
        # e.g. 看不懂
        if len(word) == 3 and word[1] == "不":
            finals[1] = finals[1][:-1] + "5"
        else:
            for i, char in enumerate(word):
                # "不" before tone4 should be bu2, e.g. 不怕
                if char == "不" and i + 1 < len(word) and finals[i + 1][-1] == "4":
                    finals[i] = finals[i][:-1] + "2"
        return finals

    def _yi_sandhi(self, word: str, finals: List[str]) -> List[str]:
        # "一" in number sequences, e.g. 一零零, 二一零
        if word.find("一") != -1 and all(
            [item.isnumeric() for item in word if item != "一"]
        ):
            return finals
        # "一" between reduplication words shold be yi5, e.g. 看一看
        elif len(word) == 3 and word[1] == "一" and word[0] == word[-1]:
            finals[1] = finals[1][:-1] + "5"
        # when "一" is ordinal word, it should be yi1
        elif word.startswith("第一"):
            finals[1] = finals[1][:-1] + "1"
        else:
            for i, char in enumerate(word):
                if char == "一" and i + 1 < len(word):
                    # "一" before tone4 should be yi2, e.g. 一段
                    if finals[i + 1][-1] == "4":
                        finals[i] = finals[i][:-1] + "2"
                    # "一" before non-tone4 should be yi4, e.g. 一天
                    else:
                        # "一" 后面如果是标点，还读一声
                        if word[i + 1] not in self.punc:
                            finals[i] = finals[i][:-1] + "4"
        return finals

    def _split_word(self, word: str) -> List[str]:
        word_list = jieba.cut_for_search(word)
        word_list = sorted(word_list, key=lambda i: len(i), reverse=False)
        first_subword = word_list[0]
        first_begin_idx = word.find(first_subword)
        if first_begin_idx == 0:
            second_subword = word[len(first_subword) :]
            new_word_list = [first_subword, second_subword]
        else:
            second_subword = word[: -len(first_subword)]
            new_word_list = [second_subword, first_subword]
        return new_word_list

    def _three_sandhi(self, word: str, finals: List[str]) -> List[str]:
        if len(word) == 2 and self._all_tone_three(finals):
            finals[0] = finals[0][:-1] + "2"
        elif len(word) == 3:
            word_list = self._split_word(word)
            if self._all_tone_three(finals):
                #  disyllabic + monosyllabic, e.g. 蒙古/包
                if len(word_list[0]) == 2:
                    finals[0] = finals[0][:-1] + "2"
                    finals[1] = finals[1][:-1] + "2"
                #  monosyllabic + disyllabic, e.g. 纸/老虎
                elif len(word_list[0]) == 1:
                    finals[1] = finals[1][:-1] + "2"
            else:
                finals_list = [finals[: len(word_list[0])], finals[len(word_list[0]) :]]
                if len(finals_list) == 2:
                    for i, sub in enumerate(finals_list):
                        # e.g. 所有/人
                        if self._all_tone_three(sub) and len(sub) == 2:
                            finals_list[i][0] = finals_list[i][0][:-1] + "2"
                        # e.g. 好/喜欢
                        elif (
                            i == 1
                            and not self._all_tone_three(sub)
                            and finals_list[i][0][-1] == "3"
                            and finals_list[0][-1][-1] == "3"
                        ):
                            finals_list[0][-1] = finals_list[0][-1][:-1] + "2"
                        finals = sum(finals_list, [])
        # split idiom into two words who's length is 2
        elif len(word) == 4:
            finals_list = [finals[:2], finals[2:]]
            finals = []
            for sub in finals_list:
                if self._all_tone_three(sub):
                    sub[0] = sub[0][:-1] + "2"
                finals += sub

        return finals

    def _all_tone_three(self, finals: List[str]) -> bool:
        return all(x[-1] == "3" for x in finals)

    # merge "不" and the word behind it
    # if don't merge, "不" sometimes appears alone according to jieba, which may occur sandhi error
    def _merge_bu(self, seg: List[Tuple[str, str]]) -> List[Tuple[str, str]]:
        new_seg = []
        last_word = ""
        for word, pos in seg:
            if last_word == "不":
                word = last_word + word
            if word != "不":
                new_seg.append((word, pos))
            last_word = word[:]
        if last_word == "不":
            new_seg.append((last_word, "d"))
            last_word = ""
        return new_seg

    # function 1: merge "一" and reduplication words in it's left and right, e.g. "听","一","听" ->"听一听"
    # function 2: merge single  "一" and the word behind it
    # if don't merge, "一" sometimes appears alone according to jieba, which may occur sandhi error
    # e.g.
    # input seg: [('听', 'v'), ('一', 'm'), ('听', 'v')]
    # output seg: [['听一听', 'v']]
    def _merge_yi(self, seg: List[Tuple[str, str]]) -> List[Tuple[str, str]]:
        new_seg = []
        # function 1
        for i, (word, pos) in enumerate(seg):
            if (
                i - 1 >= 0
                and word == "一"
                and i + 1 < len(seg)
                and seg[i - 1][0] == seg[i + 1][0]
                and seg[i - 1][1] == "v"
                and seg[i + 1][1] == "v"
            ):
                new_seg[i - 1][0] = new_seg[i - 1][0] + "一" + new_seg[i - 1][0]
            else:
                if (
                    i - 2 >= 0
                    and seg[i - 1][0] == "一"
                    and seg[i - 2][0] == word
                    and pos == "v"
                    and seg[i - 2][1] == "v"
                ):
                    continue
                else:
                    new_seg.append([word, pos])
        seg = new_seg
        new_seg = []
        # function 2
        for i, (word, pos) in enumerate(seg):
            if new_seg and new_seg[-1][0] == "一":
                new_seg[-1][0] = new_seg[-1][0] + word
            else:
                new_seg.append([word, pos])
        return new_seg

    # the first and the second words are all_tone_three
    def _merge_continuous_three_tones(
        self, seg: List[Tuple[str, str]]
    ) -> List[Tuple[str, str]]:
        new_seg = []
        sub_finals_list = [
            lazy_pinyin(word, neutral_tone_with_five=True, style=Style.FINALS_TONE3)
            for (word, pos) in seg
        ]
        assert len(sub_finals_list) == len(seg)
        merge_last = [False] * len(seg)
        for i, (word, pos) in enumerate(seg):
            if (
                i - 1 >= 0
                and self._all_tone_three(sub_finals_list[i - 1])
                and self._all_tone_three(sub_finals_list[i])
                and not merge_last[i - 1]
            ):
                # if the last word is reduplication, not merge, because reduplication need to be _neural_sandhi
                if (
                    not self._is_reduplication(seg[i - 1][0])
                    and len(seg[i - 1][0]) + len(seg[i][0]) <= 3
                ):
                    new_seg[-1][0] = new_seg[-1][0] + seg[i][0]
                    merge_last[i] = True
                else:
                    new_seg.append([word, pos])
            else:
                new_seg.append([word, pos])

        return new_seg

    def _is_reduplication(self, word: str) -> bool:
        return len(word) == 2 and word[0] == word[1]

    # the last char of first word and the first char of second word is tone_three
    def _merge_continuous_three_tones_2(
        self, seg: List[Tuple[str, str]]
    ) -> List[Tuple[str, str]]:
        new_seg = []
        sub_finals_list = [
            lazy_pinyin(word, neutral_tone_with_five=True, style=Style.FINALS_TONE3)
            for (word, pos) in seg
        ]
        assert len(sub_finals_list) == len(seg)
        merge_last = [False] * len(seg)
        for i, (word, pos) in enumerate(seg):
            if (
                i - 1 >= 0
                and sub_finals_list[i - 1][-1][-1] == "3"
                and sub_finals_list[i][0][-1] == "3"
                and not merge_last[i - 1]
            ):
                # if the last word is reduplication, not merge, because reduplication need to be _neural_sandhi
                if (
                    not self._is_reduplication(seg[i - 1][0])
                    and len(seg[i - 1][0]) + len(seg[i][0]) <= 3
                ):
                    new_seg[-1][0] = new_seg[-1][0] + seg[i][0]
                    merge_last[i] = True
                else:
                    new_seg.append([word, pos])
            else:
                new_seg.append([word, pos])
        return new_seg

    def _merge_er(self, seg: List[Tuple[str, str]]) -> List[Tuple[str, str]]:
        new_seg = []
        for i, (word, pos) in enumerate(seg):
            if i - 1 >= 0 and word == "儿" and seg[i - 1][0] != "#":
                new_seg[-1][0] = new_seg[-1][0] + seg[i][0]
            else:
                new_seg.append([word, pos])
        return new_seg

    def _merge_reduplication(self, seg: List[Tuple[str, str]]) -> List[Tuple[str, str]]:
        new_seg = []
        for i, (word, pos) in enumerate(seg):
            if new_seg and word == new_seg[-1][0]:
                new_seg[-1][0] = new_seg[-1][0] + seg[i][0]
            else:
                new_seg.append([word, pos])
        return new_seg

    def pre_merge_for_modify(self, seg: List[Tuple[str, str]]) -> List[Tuple[str, str]]:
        seg = self._merge_bu(seg)
        try:
            seg = self._merge_yi(seg)
        except:
            print("_merge_yi failed")
        seg = self._merge_reduplication(seg)
        try:
            seg = self._merge_continuous_three_tones(seg)
        except:
            print("_merge_continuous_three_tones failed")
        try:
            seg = self._merge_continuous_three_tones_2(seg)
        except:
            print("_merge_continuous_three_tones_2 failed")

        seg = self._merge_er(seg)
        return seg

    def modified_tone(self, word: str, pos: str, finals: List[str]) -> List[str]:
        finals = self._bu_sandhi(word, finals)
        finals = self._yi_sandhi(word, finals)
        finals = self._neural_sandhi(word, pos, finals)
        finals = self._three_sandhi(word, finals)
        return finals



from fastapi import APIRouter, Request, HTTPException
from sse_starlette.sse import EventSourceResponse
from ..api_response.chat_response import (
    chat_completion_response_stream,
    chat_completion_response,
    completion_response,
    completion_response_stream,
    chat_completion_response_artifact_stream,
chat_completion_response_artifact
)
from ..dependencies import get_model_manager
from ..logger import logger
from gallama.data_classes import (
    ChatMLQuery,
    ToolForce,
    GenerateQuery,
    GenQueue,
    GenQueueDynamic
)
import asyncio

# https://platform.openai.com/docs/api-reference/chat/create

router = APIRouter(prefix="/v1", tags=["chat"])


def validate_api_request(query: ChatMLQuery):
    # fixing api call parameter
    # update temperature to 0.1 because huggingface doesn't accept 0 temperature
    if query.temperature <= 0:
        query.temperature = 0.01
    # tool_choice is set to 'auto' if there is tool

    # set tool_choice if user didn't set it and ensure parameters are correct
    # if no tool, set tool_choice to none
    if query.tools:
        if query.tool_choice is None:
            if query.tools:
                query.tool_choice = "auto"
            else:
                query.tool_choice = "none"
        elif query.tool_choice == "none":
            # wipe out all the tools if any
            query.tools = None
        elif isinstance(query.tool_choice, ToolForce):
            # wipe out all the tool except the tool to be forced 'tool_choice' in api call
            force_single_tool = []
            for tool in query.tools:
                if tool.function.name == query.tool_choice.function.name:
                    force_single_tool.append(tool)

            if len(force_single_tool) != 1:
                raise Exception("tool_choice not exist in function list")
            else:
                query.tool_choice = "required"
                query.tools = force_single_tool
    else:
        query.tool_choice = "none"

    # validate that prefix_strings and regex_prefix_pattern  or regex_pattern can not be used together
    if query.prefix_strings and (query.regex_pattern or query.regex_prefix_pattern):
        raise HTTPException(status_code=400, detail="refix_strings and regex_pattern, regex_prefix_pattern can not be used together")

    return query


@router.post("/chat/completions")
async def chat_completion(request: Request, query: ChatMLQuery):

    model_manager = get_model_manager()
    gen_queue = GenQueueDynamic()      # this queue will hold the result for this generation

    # validate and fix query
    query = validate_api_request(query)

    try:
        llm = model_manager.get_model(query.model, _type="llm")
        # llm = model_manager.llm_dict.get(query.model)
        # if not llm:
        #     llm = model_manager.llm_dict[list(model_manager.llm_dict.keys())[0]]

        # log if thinking is used
        if query.thinking_template:
            logger.info(f"thinking is used with returnThinking set to {query.return_thinking}")

        # start the generation task
        asyncio.create_task(
            llm.chat(
                query=query,
                prompt_eng=llm.prompt_eng,
                gen_queue=gen_queue,
                request=request,
            )
        )

        # send the response to client
        if query.stream:
            # EventSourceResponse take iterator so need to handle at here
            if query.artifact == "No":     # not using artefact
                return EventSourceResponse(
                    chat_completion_response_stream(
                        query=query, gen_queue=gen_queue, model_name=llm.model_name, request=request,
                    ))
            else:
                return EventSourceResponse(
                    chat_completion_response_artifact_stream(
                        query=query, gen_queue=gen_queue, model_name=llm.model_name, request=request,
                    ))
        else:
            if query.artifact == "No":     # not using artefact
                return await chat_completion_response(query=query, gen_queue=gen_queue, model_name=llm.model_name, request=request,)
            else:
                return await chat_completion_response_artifact(query=query, gen_queue=gen_queue, model_name=llm.model_name, request=request,)
    except HTTPException as e:
        logger.error(e)
        return e


@router.post("/chat/generate")
@router.post("/completions")
async def generate(request: Request, query: GenerateQuery):
    model_manager = get_model_manager()
    gen_queue = GenQueueDynamic()      # this queue will hold the result for this generation
    llm = model_manager.get_model(query.model, _type="llm")
    # llm = model_manager.llm_dict[query.model]

    # start the generation task
    asyncio.create_task(llm.chat_raw(
        prompt=query.prompt,
        max_tokens=query.max_tokens,
        gen_queue=gen_queue,
        request=request,
    ))

    if query.stream:
        # EventSourceResponse take iterator so need to handle iterator here
        return EventSourceResponse(completion_response_stream(request, gen_queue, model_name=llm.model_name, request=request,))
    else:
        return await completion_response(gen_queue, model_name=llm.model_name, request=request,)


from dataclasses import dataclass
import numpy as np
from scipy import signal


@dataclass
class PreprocessingConfig:
    """Configuration for audio preprocessing features"""
    enable_highpass: bool = True
    enable_compression: bool = True
    highpass_cutoff: float = 100  # Cutoff frequency in Hz
    compression_threshold: float = -20  # dB
    compression_ratio: float = 4.0
    attack_time: float = 0.005  # seconds
    release_time: float = 0.1  # seconds
    makeup_gain: float = 6.0  # dB


class AudioPreprocessor:
    def __init__(self, config: PreprocessingConfig, sample_rate: int = 24000):
        """
        Initialize audio preprocessor for float array processing.

        Args:
            config: PreprocessingConfig object containing processing parameters
            sample_rate: Input audio sample rate in Hz (default 24000)
                Note: This should match the input audio sample rate, not the VAD rate
        """
        self.config = config
        self.sample_rate = sample_rate

        # Initialize high-pass filter if enabled
        if self.config.enable_highpass:
            nyquist = self.sample_rate / 2
            normalized_cutoff = self.config.highpass_cutoff / nyquist
            self.hp_b, self.hp_a = signal.butter(4, normalized_cutoff, btype='high')
            self.hp_zi = None  # Filter state for high-pass

        # Initialize compressor parameters if enabled
        if self.config.enable_compression:
            # Convert from dB to linear
            self.threshold = 10 ** (self.config.compression_threshold / 20)
            self.makeup_gain = 10 ** (self.config.makeup_gain / 20)

            # Time constants
            self.attack = np.exp(-1 / (self.sample_rate * self.config.attack_time))
            self.release = np.exp(-1 / (self.sample_rate * self.config.release_time))

            # Envelope follower state
            self.env = 0.0

    def _apply_highpass(self, audio: np.ndarray) -> np.ndarray:
        """
        Apply high-pass filter to audio.

        Args:
            audio: Input audio as float32 numpy array in range [-1.0, 1.0]

        Returns:
            Filtered audio as float32 numpy array
        """
        if self.hp_zi is None:
            # Initialize filter state
            self.hp_zi = signal.lfilter_zi(self.hp_b, self.hp_a)

        # Apply filter and update state
        filtered, self.hp_zi = signal.lfilter(
            self.hp_b, self.hp_a, audio, zi=self.hp_zi * audio[0]
        )
        return filtered

    def _apply_compression(self, audio: np.ndarray) -> np.ndarray:
        """
        Apply dynamic range compression.

        Args:
            audio: Input audio as float32 numpy array in range [-1.0, 1.0]

        Returns:
            Compressed audio as float32 numpy array
        """
        output = np.zeros_like(audio)
        env = self.env

        for i, sample in enumerate(audio):
            # Envelope detection
            level = abs(sample)
            if level > env:
                env = self.attack * env + (1 - self.attack) * level
            else:
                env = self.release * env + (1 - self.release) * level

            # Gain computation
            if env > self.threshold:
                gain_reduction = self.threshold + (env - self.threshold) / self.config.compression_ratio
                gain = gain_reduction / env if env > 0 else 1.0
            else:
                gain = 1.0

            # Apply gain and makeup gain
            output[i] = sample * gain * self.makeup_gain

        # Update envelope state
        self.env = env
        return output

    def process_float_chunk(self, audio_float: np.ndarray) -> np.ndarray:
        """
        Process an audio chunk in float format.

        Args:
            audio_float: Input audio as float32 numpy array in range [-1.0, 1.0]
                        at input_sample_rate (24000 Hz by default)

        Returns:
            Processed audio as float32 numpy array in range [-1.0, 1.0]
            at the same sample rate as input
        """
        if audio_float.size == 0:
            return audio_float

        try:
            processed_audio = audio_float

            if self.config.enable_highpass:
                processed_audio = self._apply_highpass(processed_audio)

            if self.config.enable_compression:
                processed_audio = self._apply_compression(processed_audio)

            # Ensure output is within [-1, 1] range
            return np.clip(processed_audio, -1.0, 1.0)

        except Exception as e:
            print(f"Error in audio preprocessing: {str(e)}")
            return audio_float

    def reset(self):
        """Reset all internal states"""
        if self.config.enable_highpass:
            self.hp_zi = None
        if self.config.enable_compression:
            self.env = 0.0



from .thinking_template import THINKING_TEMPLATE

# selective import as not all backend available on all platform
try:
    from .engine import ModelExllama
except ImportError:
    ModelExllama = None

try:
    from .engine import ModelLlamaCpp
except ImportError:
    ModelLlamaCpp = None

try:
    from .engine import ModelTransformers
except ImportError:
    ModelTransformers = None

# try:
from .engine import ModelMLXVLM
# except ImportError:
#     ModelMLXVLM = None


#!/usr/bin/env python3
"""Initialize modules for espnet2 neural networks."""
import torch
from typeguard import check_argument_types


def initialize(model: torch.nn.Module, init: str):
    """Initialize weights of a neural network module.

    Parameters are initialized using the given method or distribution.

    Custom initialization routines can be implemented into submodules
    as function `espnet_initialization_fn` within the custom module.

    Args:
        model: Target.
        init: Method of initialization.
    """
    assert check_argument_types()
    print("init with", init)

    # weight init
    for p in model.parameters():
        if p.dim() > 1:
            if init == "xavier_uniform":
                torch.nn.init.xavier_uniform_(p.data)
            elif init == "xavier_normal":
                torch.nn.init.xavier_normal_(p.data)
            elif init == "kaiming_uniform":
                torch.nn.init.kaiming_uniform_(p.data, nonlinearity="relu")
            elif init == "kaiming_normal":
                torch.nn.init.kaiming_normal_(p.data, nonlinearity="relu")
            else:
                raise ValueError("Unknown initialization: " + init)
    # bias init
    for name, p in model.named_parameters():
        if ".bias" in name and p.dim() == 1:
            p.data.zero_()



try:
    from .faster_whisper.model import ASRFasterWhisper
except ModuleNotFoundError:
    ASRFasterWhisper = None

try:
    from .mlx_whisper.model import ASRMLXWhisper
except ModuleNotFoundError:
    ASRMLXWhisper = None


from .asr_processor import ASRProcessor


from ...base import TTSBase
import os
from kokoro import KPipeline
from kokoro.pipeline import LANG_CODES
from kokoro.model import KModel
import torch
from gallama.data_classes import ModelSpec
from gallama.logger import logger
import asyncio
from typing import Dict, Any, AsyncGenerator, Tuple, AsyncIterator, List
import numpy as np
from concurrent.futures import ThreadPoolExecutor
import time
from split_lang import LangSplitter
from split_lang.model import SubString

from huggingface_hub import hf_hub_download


class KPipelineModified(KPipeline):
    """
    KPipeline class with modify method that not use another path that can work offline
    """

    def __init__(self, *args, **kwargs):
        self.voice_path = kwargs.pop("model_path", None)    # extra argument for path of the voices pt
        super().__init__(*args, **kwargs)


    def load_single_voice(self, voice: str):
        if voice in self.voices:
            return self.voices[voice]
        if voice.endswith('.pt'):
            f = voice
        else:

            # modification from here
            use_local_voice = False
            try:
                logger.info(f"Attempt to use voice from: {self.voice_path}/voices/{voice}.pt")
                if self.voice_path and os.path.exists(f"{self.voice_path}/voices/{voice}.pt"):
                    f = f"{self.voice_path}/voices/{voice}.pt"
                    use_local_voice = True
                else:
                    raise FileNotFoundError
            except Exception as e:
                logger.info(f"Error when loading voice {voice} from local: {e}. Fall back to download from huggingface")

            # fall back to download from hugging face
            if not use_local_voice:
                f = hf_hub_download(repo_id=KModel.REPO_ID, filename=f'voices/{voice}.pt')

            if not voice.startswith(self.lang_code):
                v = LANG_CODES.get(voice, voice)
                p = LANG_CODES.get(self.lang_code, self.lang_code)
                logger.warning(f'Language mismatch, loading {v} voice into {p} pipeline.')


        pack = torch.load(f, weights_only=True)
        self.voices[voice] = pack
        return pack


class TTSKokoro(TTSBase):
    def __init__(self, model_spec: ModelSpec):
        super().__init__(model_spec)
        self.executor = ThreadPoolExecutor(max_workers=1)
        self.sample_rate = 24000

    def convert_language_code(self, lan:str) -> str:
        if lan=="en":
            return "a"
        elif lan=="en-b":
            return "b"
        elif lan=="zh":
            return "z"
        elif lan=="ja":
            return "j"
        elif lan=="fr":
            return "f"
        elif lan=="hi":
            return "h"
        elif lan=="it":
            return "i"
        elif lan=="pt":
            return "p"
        else:
            logger.error(f"Unknown language code {lan}")
            return "a"  # return english

    @property
    def languages(self):
        return ["zh", "en", "en-b", "ja", "fr", "hi", "it", "pt"]

    def load_model(self, model_spec: ModelSpec):

        self.model = {}

        # create English pipeline:
        model_object = None

        for lang_code in self.languages:
            kokoro_lang_code = self.convert_language_code(lang_code)
            if model_object is None:
                pipeline = KPipelineModified(lang_code=kokoro_lang_code, model_path=model_spec.model_id)
                model_object = pipeline.model
            else:
                pipeline = KPipelineModified(lang_code=kokoro_lang_code, model=model_object, model_path=model_spec.model_id)

            self.model[lang_code] = pipeline

    async def stream_audio(
        self,
        inputs: Dict[str, Any],
        queue: asyncio.Queue,
        session_tracker: set = None,  # Set to track generations for a specific session
    ) -> None:
        """
        Feeds audio fragments directly into the provided queue for upstream consumption.

        Args:
            inputs: Dictionary of TTS parameters including text, ref_audio_path, etc.
            queue: asyncio.Queue where audio chunks will be placed
            session_tracker: Optional set to track generations for this specific session
        """

        task_id = str(time.time())
        if session_tracker is not None:
            session_tracker.add(task_id)

        loop = asyncio.get_running_loop()

        pipeline_to_use: KPipeline = self.model.get(inputs['text_lang'], self.model['en'])  # load en if language not found

        # remove language field as it is not part of KPipeline __call__ argument
        inputs.pop('text_lang')

        def run_tts():
            try:
                for gs, ps, audio_chunk in pipeline_to_use(**inputs):
                    # gs => graphemes/text
                    # ps => phonemes
                    future = asyncio.run_coroutine_threadsafe(
                        queue.put((self.sample_rate, audio_chunk.numpy())),     # kokoro return torch.Tensor, hence convert to numpy
                        loop
                    )
                    future.result()

            except Exception as e:
                logger.error(f"Error in TTS generation {task_id}: {e}")
                future = asyncio.run_coroutine_threadsafe(
                    queue.put(Exception(f"TTS error: {str(e)}")),
                    loop
                )
                future.result()
            finally:
                # Remove this task from session tracker if it exists
                if session_tracker is not None:
                    session_tracker.discard(task_id)

        loop.run_in_executor(self.executor, run_tts)

    def check_language(self, text_list: List[SubString]):
        supported_languages = self.languages

        for text in text_list:
            if text.lang not in supported_languages:
                text.lang = "en"

        if not text_list:
            return text_list

        grouped_list = [text_list[0]]
        for current in text_list[1:]:
            last = grouped_list[-1]
            # If current and last have the same language, merge them.
            if last.lang == current.lang:
                last.text += current.text
                last.length += current.length
            else:
                grouped_list.append(current)

        return grouped_list


    async def text_to_speech(
        self,
        text: str,
        language: str = "auto",
        stream: bool = False,
        batching: bool = False,
        batch_size: int = 1,
        queue: asyncio.Queue = None,
        voice: str = None,
        speed_factor: float = None,
        session_tracker: set = None,
        **kwargs: Any
    ) -> None | Tuple[int, np.ndarray] | AsyncIterator[Tuple[int, np.ndarray]] | str:


        if language == "auto":
            DEFAULT_LANG = "en"
            lang_splitter = LangSplitter(
                merge_across_punctuation=True
            )

            text_segmented_by_lan = self.check_language(
                lang_splitter.split_by_lang(text)
            )


            if len(text_segmented_by_lan) == 1:
                logger.debug(f"single language: {text_segmented_by_lan[0].lang}")
                return await self.text_to_speech_single_language(
                    text,
                    text_segmented_by_lan[0].lang,   # the detected language
                    stream,
                    batching,
                    batch_size,
                    queue,
                    voice,
                    speed_factor,
                    session_tracker,
                    ** kwargs
                )
            else:
                # example of each item in text_segmented_by_lan:
                # {'lang': 'zh', 'text': '我喜欢在雨天里听音乐。\n', 'score': 1.0}
                logger.info(f"Processing multi-language text with {len(text_segmented_by_lan)} segments")

                for text_chunk in text_segmented_by_lan:
                    _lang = text_chunk.lang
                    _text = text_chunk.text
                    # Multiple languages case

                    if stream and not queue:
                        raise ValueError("For streaming mode, a queue must be provided")

                    if not stream:
                        # Non-stream: Aggregate results from segments
                        full_audio = []
                        sample_rate = None

                        for segment in text_segmented_by_lan:
                            lang = segment.lang
                            seg_text = segment.text
                            if not seg_text.strip():
                                continue

                            # Process each segment with stream=False
                            sr, audio = await self.text_to_speech_single_language(
                                seg_text,
                                lang,
                                stream=False,
                                batching=batching,
                                batch_size=batch_size,
                                queue=None,
                                voice=voice,
                                speed_factor=speed_factor,
                                session_tracker=session_tracker,
                                **kwargs
                            )

                            if sample_rate is None:
                                sample_rate = sr
                            elif sr != sample_rate:
                                raise ValueError(f"Sample rate mismatch between segments: {sr} vs {sample_rate}")

                            full_audio.append(audio)

                        if not full_audio:
                            return sample_rate or self.sample_rate, np.array([], dtype=np.int16)

                        return sample_rate, np.concatenate(full_audio)
                    else:
                        # Stream: Process segments sequentially using shared queue
                        for segment in text_segmented_by_lan:
                            lang = segment.lang
                            seg_text = segment.text
                            if not seg_text.strip():
                                continue

                            # Wait until all previous tasks are completed
                            while session_tracker:
                                await asyncio.sleep(0.1)

                            await self.text_to_speech_single_language(
                                seg_text,
                                lang,
                                stream=True,
                                batching=batching,
                                batch_size=batch_size,
                                queue=queue,
                                voice=voice,
                                speed_factor=speed_factor,
                                session_tracker=session_tracker,
                                **kwargs
                            )
                        return None
        else:

            return await self.text_to_speech_single_language(
                text,
                language,
                stream,
                batching,
                batch_size,
                queue,
                voice,
                speed_factor,
                session_tracker,
                ** kwargs
            )


    async def text_to_speech_single_language(
        self,
        text: str,
        language: str = "auto",
        stream: bool = False,
        batching: bool = False,
        batch_size: int = 1,
        queue: asyncio.Queue = None,
        voice: str = None,
        speed_factor: float = None,
        session_tracker: set = None,
        **kwargs: Any
    ) -> None | Tuple[int, np.ndarray] | AsyncIterator[Tuple[int, np.ndarray]] | str:
        """
        Generate audio chunks from text and put them into an asyncio Queue.

        Args:
            text: Text to convert to speech
            language: Language of the text (default: "auto")
            stream: Whether to stream the audio in chunks (default: False)
            voice: the name of the voice to use (default: None), if None, will use default voice
            speed_factor: Speed factor for the audio playback (default: 1.0)
            batching: whether to using parallel batching, will be faster but require more memory (default: False)
            batch_size: batch size if use batching (default: 1)
            queue: Optional asyncio Queue to receive audio chunks
            session_tracker: Optional set to track generations for this specific session
            kwargs: Additional parameters to pass to the text_to_speech function

        Returns:
            If stream=False: Tuple of (sample_rate, concatenated_audio_data)
            If stream=True: None (audio chunks are sent to the provided queue)
        """
        logger.debug(f"Converting TTS: {text}")

        if stream and not queue:
            # for streaming, the result will be written directly to the queue hence it is required parameter
            raise Exception("For streaming mode, the queue must be provided")

        if voice:
            voice_to_use = voice
        else:
            voice_to_use = "af_heart"
            speed_factor = 1

        if speed_factor:
            speed_factor_to_use = speed_factor
        else:
            raise Exception("There must be at least one voice to use")

        params = {
            "text": text,
            "text_lang": language,
            "voice": voice_to_use,
            "speed": speed_factor_to_use,
        }

        if stream:
            if queue is None:
                queue = asyncio.Queue()
            await self.stream_audio(
                inputs=params,
                queue=queue,
                session_tracker=session_tracker
            )
            return None  # Since we're using a queue, no direct return value needed
        else:
            # For non-streaming mode, create a temporary queue and collect all chunks
            temp_queue = asyncio.Queue()
            temp_session_tracker = set()  # Create temporary session tracker if none provided
            session_tracker_to_use = session_tracker if session_tracker is not None else temp_session_tracker

            await self.stream_audio(
                inputs=params,
                queue=temp_queue,
                session_tracker=session_tracker_to_use
            )

            try:
                # Collect all audio chunks into a single array
                audio_chunks = []
                sample_rate = None

                # Process chunks while the session is active or queue has items
                while session_tracker_to_use or not temp_queue.empty():
                    try:
                        # Use asyncio.wait_for with a timeout for the queue.get()
                        chunk = await asyncio.wait_for(temp_queue.get(), timeout=0.1)
                        if isinstance(chunk, Exception):
                            raise chunk

                        current_sample_rate, audio_data = chunk
                        if sample_rate is None:
                            sample_rate = current_sample_rate
                        elif sample_rate != current_sample_rate:
                            raise ValueError(
                                f"Inconsistent sample rates detected: {sample_rate} vs {current_sample_rate}")

                        if audio_data.shape[0] > 0:  # Only add non-empty chunks
                            audio_chunks.append(audio_data)
                        temp_queue.task_done()
                    except asyncio.TimeoutError:
                        # Timeout is expected when queue is empty, just continue the loop
                        continue

                if not audio_chunks:
                    return 0, np.array([], dtype=np.int16)

                return sample_rate, np.concatenate(audio_chunks)

            except Exception as e:
                logger.error(f"Error during audio generation: {e}")
                # self.model.stop()
                raise


import argparse
import shutil
from pathlib import Path
from gallama.config import ConfigManager
from gallama.logger.logger import logger
from gallama.server import run_from_script
from gallama.server_routes import download_model
from gallama.data_classes.data_class import ModelDownloadSpec, SUPPORTED_BACKENDS
from rich.markdown import Markdown
from rich.console import Console


def get_data_dir() -> Path:
    """Get the absolute path to the data directory."""
    return Path(__file__).parent / 'data'

def get_data_dir_dev() -> Path:
    """Get the absolute path to the data directory."""
    return Path(__file__).parent / 'gallama' / 'data'


def find_config_file(config_dir, filename):
    """Find the configuration file with either .yml or .yaml extension."""
    yml_file = config_dir / f"{filename}.yml"
    yaml_file = config_dir / f"{filename}.yaml"
    logger.debug(f"Searching for config files: {yml_file}, {yaml_file}")
    return yml_file if yml_file.exists() else yaml_file if yaml_file.exists() else None


def ensure_config_file():
    """Ensure the user-specific configuration file exists."""
    home_dir = Path.home()
    config_dir = home_dir / "gallama"
    model_dir = home_dir / "gallama" / "models"
    config_filename = "model_config"

    config_dir_default = get_data_dir()
    config_dir_default_dev = get_data_dir_dev()     # dev mode file path is different

    try:
        if not config_dir.exists():
            config_dir.mkdir(parents=True)
            logger.info(f"Created configuration directory: {config_dir}")

        if not model_dir.exists():
            model_dir.mkdir(parents=True)
            logger.info(f"Created configuration directory: {model_dir}")

        config_file = find_config_file(config_dir, config_filename)
        if not config_file:

            default_config = find_config_file(config_dir_default, config_filename)
            default_config_dev = find_config_file(config_dir_default_dev, config_filename)

            if default_config:
                config_file = config_dir / default_config.name
                shutil.copy(default_config, config_file)
                logger.info(f"Created default configuration file: {config_file}")
            elif default_config_dev:
                config_file = config_dir / default_config_dev.name
                shutil.copy(default_config_dev, config_file)
                logger.info(f"Created default configuration file: {default_config_dev}")
            else:
                raise FileNotFoundError(f"Default configuration file not found in: {config_dir}")

        return config_file
    except Exception as e:
        logger.error(f"Error creating configuration: {str(e)}")
        logger.info("Please ensure the default configuration file exists in the package's data directory.")
        return None


def parse_dict(arg):
    """Parses a key=value string and returns a dictionary."""
    result = {}
    for pair in arg.split():

        key, value = pair.split('=')
        result[key] = value.strip("'")  # Strip single quotes here as well
    return result


def run_server(host, port):
    import uvicorn
    from .app import app
    uvicorn.run(app, host=host, port=port)


def main_cli():
    arg_parser = argparse.ArgumentParser(description="Launch multi model src instance")
    arg_parser.add_argument('-v', "--verbose", action='store_true', help="Turn on more verbose logging")

    subparsers = arg_parser.add_subparsers(dest="command")

    # Add 'run' subcommand
    serve_parser = subparsers.add_parser("run", help="Run the FastAPI server_engine")
    serve_parser.add_argument("model_name", nargs='?', help="Model name to run (simplified version)")
    serve_parser.add_argument("--strict_mode", action="store_true", default=False,
                              help="Enable strict mode for routing non-embedding requests to matching model names")
    serve_parser.add_argument("-id", "--model_id", action='append', type=parse_dict, default=None,
                              help="Model configuration. Can be used multiple times for different models. "
                                   "Format: -id model_id=<model_name> [gpu=<vram1>,<vram2>,...] [gpu<n>=<vram>] [cache=<size>] [model_type=exllama|embedding]"
                                   "Examples:\n"
                                   "1. Specify all GPUs: -id model_id=/path/llama3-8b gpu=4,2,0,3 cache=8192\n"
                                   "2. Specify individual GPUs: -id model_id=/path/llama3-8b gpu0=8 gpu1=8 cache=8192\n"
                                   "3. Mix formats: -id model_id=gpt4 gpu=4,2 gpu3=3 cache=8192\n"
                                   "4. Embedding model: -id model_id=/path/gte-large-en-v1.5 gpu=0,0,0,3"
                                   "cache_size is defaulted to model context length if not specified"
                                   "cache_size can be more than context length, which will help model perform better in batched generation"
                                   "cache_size is not application to embedding model"
                                   "VRAM is specified in GB. Cache size is integer which is the context length to cache."
                                   "VRAM for embedding will simple set env parameter to allow infinity_embedding to view the specific GPU and can not enforce VRAM size restriction")
    serve_parser.add_argument("--host", type=str, default="127.0.0.1", help="The host to bind to.")
    serve_parser.add_argument('-p', "--port", type=int, default=8000, help="The port to bind to.")
    serve_parser.add_argument('-v', "--verbose", action='store_true', help="Turn on more verbose logging")


    # Add 'download' subcommand
    download_parser = subparsers.add_parser("download", help="Download a model from Hugging Face")
    download_parser.add_argument("model_spec", type=str, help="Model specification in the format 'model_name:quant'")
    download_parser.add_argument("--backend", type=str, default=None, choices=SUPPORTED_BACKENDS,
                                 help="The backend to download model with. One of exllama, llama_cpp, embedding or transformers")

    # Modify 'list' subcommand
    list_parser = subparsers.add_parser("list", help="List models")
    list_parser.add_argument("type", nargs='?', choices=['available', 'downloaded'], default='downloaded',
                             help="List downloaded model (default) or list available to show the list of model that can be downloaded")

    args = arg_parser.parse_args()

    # ensure config file is there
    ensure_config_file()

    # set logger level
    if args.verbose:
        logger.setLevel("DEBUG")
    else:
        logger.setLevel("INFO")

    if args.command == "run" or args.command is None:
        # if args.model_name and not args.model_id:
        #     # Convert simplified version to the full version
        #     args.model_id = [{"model_name": args.model_name}]
        run_from_script(args)  # Pass all arguments to run_from_script

    elif args.command == "download":
        if ":" in args.model_spec:
            model_name, quant = args.model_spec.split(':')
        else:
            # default download mode
            model_name = args.model_spec
            quant = None

        download_model(ModelDownloadSpec(
            model_name=model_name,
            quant=float(quant) if quant else None,
            backend=args.backend,
        ))

    elif args.command == "list":
        config_manager = ConfigManager()
        console = Console()
        if args.type == "available":
            md = Markdown(config_manager.list_available_models_table)
            console.print(md)
        elif args.type == "downloaded" or args.type is None:
            md = Markdown(config_manager.list_downloaded_models_table.lstrip())
            console.print(md)
        elif args.type == "running":
            # Implement logic to list running models
            pass


if __name__ == "__main__":
    main_cli()


# llm_server.py
from fastapi import WebSocket, WebSocketDisconnect, APIRouter
from typing import AsyncGenerator, TypeVar
from collections import OrderedDict

from pydantic.json import pydantic_encoder
from ..api_response.chat_response import chat_completion_response_stream
from ..data_classes.realtime_server_proto import ContentTypeServer, ConversationItemServer, parse_conversation_item, MessageContentServer
from ..logger import logger
from ..data_classes.realtime_client_proto import *
from ..routes.chat import validate_api_request
from ..data_classes.data_class import (
    BaseMessage,
    ChatMLQuery,
    ToolCall,
    FunctionCall,
    ChatCompletionResponse,
    ChatMessage,
    OneTool,
    ToolCallResponse,
    StreamChoice,
    MultiModalTextContent,
    MultiModalImageContent,
    UsageResponse,
    ToolSpec,
    FunctionSpec,
    ParameterSpec,
)

from ..data_classes.internal_ws import WSInterLLM, WSInterConfigUpdate

from ..data_classes.generation_data_class import (
    GenQueueDynamic,
)
from ..data_classes.realtime_client_proto import (
    SessionConfig
)
from ..data_classes.video import VideoFrameCollection


import json
import asyncio

from ..dependencies import (
    get_model_manager,
    get_video_collection,
    get_session_config,
    dependency_update_session_config,
    get_record_start_time,
    get_clear_history_flag,
    set_clear_history_flag
)


router = APIRouter(prefix="", tags=["llm"])


class LLMGenerateParams(BaseModel):
    temperature: float = 0.8
    max_tokens: Optional[int] = None


class GenerateEvent(BaseModel):
    type: Literal["generate", "generate_cache"]
    params: Optional[LLMGenerateParams] = None





T = TypeVar("T", bound=ConversationItemServer)


class WebSocketSession:
    def __init__(self, websocket: WebSocket, session_id: str):
        self.websocket = websocket
        self.session_id = session_id
        self.conversation_history = OrderedDict()
        self.session_config = SessionConfig()

        # job cancelling
        self.stop_event = asyncio.Event()

        # Start background task to monitor the clear history flag.
        self.monitor_task = asyncio.create_task(self.monitor_clear_history_flag())

    def reset(self):
        self.conversation_history = OrderedDict()

    async def monitor_clear_history_flag(self):
        """Continuously check if clear history flag is set.
        If it is, reset the conversation history and clear the flag.
        """
        while True:
            await asyncio.sleep(1)  # check every second
            if get_clear_history_flag():
                logger.info("Clear history flag detected. Resetting conversation history.")
                self.reset()
                set_clear_history_flag(False)

    async def update_session_config(self, new_session_config: SessionConfig):
        """Update session config"""
        self.session_config = self.session_config.merge(new_session_config)

        # update session config in dependencies for video to refer to
        await dependency_update_session_config(self.session_config)

    def convert_conversation_to_chatml(
        self,
        session_config: SessionConfig,
        video_start_time: float = None,
        video_end_time: float = None,
    ) -> ChatMLQuery:
        """Convert conversation history to ChatML format"""
        messages = []

        # track the number of message retained for past message
        number_of_retained_frames = 0

        total_user_messages = sum(
            1 for item in self.conversation_history.values()
            if item.type == "message" and item.content
        )
        user_message_counter = 0

        if session_config.instructions != "":
            messages.append(BaseMessage(
                role="system",
                content=session_config.instructions
            ))

        for item_id, item in self.conversation_history.items():
            role = item.role.value if hasattr(item, 'role') and item.role else "user"

            if item.type == "message" and item.content:

                user_message_counter += 1
                message_content: List[Union[MultiModalTextContent, MultiModalImageContent]] = []

                for content in item.content:
                    if content.type in [
                        ContentTypeServer.INPUT_TEXT,
                        ContentTypeServer.TEXT,
                        ContentTypeServer.INPUT_AUDIO,
                        ContentTypeServer.AUDIO
                    ]:
                        message_content.append(MultiModalTextContent(
                            type="text",
                            text=content.text or content.transcript
                        ))
                    elif content.type == "input_image":
                        # image are the type that sent by user
                        message_content.append(MultiModalImageContent(
                            type="image_url",
                            image_url=MultiModalImageContent.ImageDetail(
                                url=content.image,
                                detail="high"
                            )
                        ))
                    elif content.type == "video_frame":
                        # video frame is prepared by the websocket itself not sent by user
                        # x2 assuming one user follow by 1 assistant message
                        if user_message_counter > total_user_messages - (session_config.video.max_message_with_retained_video * 2):
                            message_content.append(MultiModalImageContent(
                                type="image_url",
                                image_url=MultiModalImageContent.ImageDetail(
                                    url=content.image,
                                    detail="high"
                                )
                            ))
                            number_of_retained_frames += 1

                messages.append(BaseMessage(
                    role=role,
                    content=message_content
                ))
            elif item.type == "function_call":
                logger.info(f"Function call: {item}")
                messages.append(BaseMessage(
                    role="assistant",
                    tool_calls=[
                        ToolCall(
                            id=item.id,
                            type="function",
                            function=FunctionCall(
                                name=item.name,
                                arguments=item.arguments
                            )
                        )
                    ]
                ))
            elif item.type == "function_call_output":
                logger.info(f"Function call output: {item}")
                messages.append(BaseMessage(
                    role="tool",
                    content=item.output,
                    tool_call_id=item.call_id
                ))

        # Convert tools from SessionConfig to ToolSpec format
        tools = None
        if session_config.tools and len(session_config.tools) > 0:
            tools = [
                ToolSpec(
                    type="function",
                    function=FunctionSpec(
                        name=tool.name,
                        description=tool.description,
                        parameters=ParameterSpec(
                            type="object",
                            properties=tool.parameters.properties,
                            required=tool.parameters.required
                        )
                    )
                ) for tool in session_config.tools
            ]

        # Handling video using the updated video settings
        video_for_llm = None
        if session_config.video.video_stream:
            record_start_time = get_record_start_time()
            logger.info("recording start time: %s", record_start_time)

            video_collection = get_video_collection()
            if video_start_time and video_end_time:
                video_for_llm = video_collection.get_frames_between_timestamps(video_start_time, video_end_time)
            else:
                video_for_llm = video_collection.get_all_frames(start_time=record_start_time)
                video_collection.clear()

            if number_of_retained_frames >0:
                logger.info(f"Number of video frames retained for past conversation: {number_of_retained_frames}")

            logger.info(f"Number of video frames used for LLM: {len(video_for_llm)}")

            if len(video_for_llm) > 0:
                # # Insert a special token indicating that video is available.
                # messages.append(BaseMessage(
                #     role="user",
                #     content="{{VIDEO-PlaceHolderTokenHere}}"
                # ))

                # Determine which video retention mode to use.
                retained_video_frames = []
                if session_config.video.retain_video == "disable":
                    retained_video_frames = []
                elif session_config.video.retain_video == "message_based":
                    retained_video_frames = VideoFrameCollection.get_retained_frames_from_list(
                        frames=video_for_llm,
                        retained_video_frames_per_message=session_config.video.retain_per_message,
                        return_base64=True
                    )
                elif session_config.video.retain_video == "time_based":
                    # New mode: use second_per_retain interval sampling.
                    sampled_frames = VideoFrameCollection.sample_frames_time_based(video_for_llm, session_config.video.second_per_retain)

                    # Convert sampled frames to base64-encoded PNG Data URIs.
                    retained_video_frames = VideoFrameCollection.convert_frames_to_base64(sampled_frames)

                # Update the last conversation item with the retained video frames if possible.
                if retained_video_frames:
                    last_item = next(reversed(self.conversation_history.values()), None)
                    if last_item is not None and last_item.type == "message" and last_item.content:
                        for base64_frame in retained_video_frames:
                            last_item.content.append(
                                MessageContentServer(
                                    type="video_frame",
                                    image=base64_frame,
                                )
                            )

        max_token_to_use = session_config.max_response_output_tokens if session_config and session_config.max_response_output_tokens != "inf" else None
        query = ChatMLQuery(
            messages=messages,
            model=session_config.model,
            temperature=session_config.temperature if session_config else 0.8,
            stream=True,
            max_tokens=max_token_to_use,
            artifact="No",
            tools=tools if tools else None,
            # Tool extra settings:
            tool_call_thinking=session_config.tool_call_thinking,
            tool_call_thinking_token=session_config.tool_call_thinking_token,
            tool_instruction_position=session_config.tool_instruction_position,
            tool_schema_position=session_config.tool_schema_position,
            video=video_for_llm if video_for_llm and len(video_for_llm) > 0 else None,
        )

        return validate_api_request(query)

    async def process_generation_stream(
        self,
        gen_queue: GenQueueDynamic,
        model_name: str,
        query: ChatMLQuery
    ) -> AsyncGenerator[dict, None]:
        """Process generation queue and yield response chunks by wrapping chat_completion_response_stream"""
        # Create a mock request object since chat_completion_response_stream expects one
        mock_request = type('MockRequest', (), {'is_disconnected': lambda: False})()

        async for chunk in chat_completion_response_stream(
            query=query,
            gen_queue=gen_queue,
            model_name=model_name,
            request=mock_request
        ):
            # Check if generation should be stopped
            if self.stop_event.is_set():
                break

            # Handle the [DONE] message
            if chunk.get("data") == "[DONE]":
                continue

            # Parse the chunk data
            data = json.loads(chunk["data"])

            # Handle usage statistics differently for websocket
            if "usage" in data:
                yield {
                    "type": "usage.update",
                    "usage": json.dumps(data)
                }
            # For normal message chunks, pass through the data
            else:
                yield data

    @staticmethod
    def update_ordered_dict(od: OrderedDict, target_key, new_value) -> OrderedDict:
        return OrderedDict((k, new_value if k == target_key else v) for k, v in od.items())

    async def handle_message(self, message: dict):
        """Handle incoming WebSocket messages for this session"""
        logger.info(f"Received message: {message}")

        if message.get("object") == "realtime.item":
            item = parse_conversation_item(message)

            if item.id not in self.conversation_history.keys():
                # new item
                self.conversation_history[item.id] = item
            else:
                # update existing item
                self.conversation_history = self.update_ordered_dict(
                    od=self.conversation_history,
                    target_key=item.id,
                    new_value=item,
                )

            await self.websocket.send_json({
                "type": "conversation.update.ack",
                "item_id": item.id
            })
        elif message.get("type") == "conversation.item.deleted":
            # Handle deletion event
            item_id = message.get("item_id")
            if item_id in self.conversation_history:
                del self.conversation_history[item_id]

            await self.websocket.send_json({
                "type": "conversation.update.ack",
                "item_id": item_id
            })

        elif message.get("type") in ["response.create"]:
            message_pydantic = ResponseCreate(**message)

            # Reset stop event before starting new generation
            self.stop_event.clear()

            gen_queue = GenQueueDynamic()
            response_config = self.session_config.merge(message.get("response", {}))
            query = self.convert_conversation_to_chatml(
                response_config,
                video_start_time = message_pydantic.speech_start_time,
                video_end_time = message_pydantic.speech_end_time
            )

            # Ensure stream is set to True for streaming responses
            query.stream = True

            model_manager = get_model_manager()
            llm = model_manager.get_model(query.model, _type="llm")
            # llm = model_manager.llm_dict.get(query.model)
            #
            # if not llm:
            #     llm = model_manager.llm_dict.get(list(model_manager.llm_dict.keys())[0])

            if llm is None:
                error_response = {
                    "type": "error",
                    "error": {
                        "code": "model_not_found",
                        "message": f"Model '{query.model}' not found or not available"
                    }
                }
                await self.websocket.send_json(error_response)
                return

            asyncio.create_task(
                llm.chat(
                    query=query,
                    prompt_eng=llm.prompt_eng,
                    gen_queue=gen_queue,
                    request=None,
                    stop_event=self.stop_event,
                )
            )

            if message["type"] == "response.create":
                async for chunk in self.process_generation_stream(gen_queue, llm.model_name, query):
                    await self.websocket.send_json(chunk)

                await self.websocket.send_json({
                    "type": "generation.complete",
                })
            # elif essage["type"] == "response.create_cache":  # generate_cache
            #     async for _ in self.process_generation_stream(gen_queue, llm.model_name, query):
            #         pass
            #     await self.websocket.send_json({
            #         "type": "pre_cache.complete",
            #     })
            else:
                raise Exception(f"Unknown message type: {message['type']}")
        elif message.get("type") == "common.cancel":
            # abort generation
            self.stop_event.set()
        elif message.get("type") == "common.cleanup":
            # abort generation
            self.reset()
        elif message.get("type") == "common.config_update":
            message_pydantic = WSInterConfigUpdate(**message)
            await self.update_session_config(message_pydantic.config)


class LLMWebSocketServer:
    def __init__(self):
        self.sessions: Dict[str, WebSocketSession] = {}

    async def handle_websocket(self, websocket: WebSocket):
        await websocket.accept()
        session_id = str(id(websocket))

        session = WebSocketSession(websocket, session_id)
        self.sessions[session_id] = session

        try:
            while True:
                message = await websocket.receive_json()
                await session.handle_message(message)
        except WebSocketDisconnect:
            self.cleanup_session(session_id)
        except Exception as e:
            logger.error(f"Error in LLM websocket: {str(e)}")
            self.cleanup_session(session_id)

    def cleanup_session(self, session_id: str):
        """Clean up session data"""
        if session_id in self.sessions:
            del self.sessions[session_id]


llm_server = LLMWebSocketServer()


@router.websocket("/llm")
async def websocket_endpoint(websocket: WebSocket):
    await llm_server.handle_websocket(websocket)


# modified from https://github.com/yangdongchao/SoundStorm/blob/master/soundstorm/s1/AR/modules/lr_schedulers.py
# reference: https://github.com/lifeiteng/vall-e
import math

import torch
from matplotlib import pyplot as plt
from torch import nn
from torch.optim import Adam


class WarmupCosineLRSchedule(torch.optim.lr_scheduler._LRScheduler):
    """
    Implements Warmup learning rate schedule until 'warmup_steps', going from 'init_lr' to 'peak_lr' for multiple optimizers.
    """

    def __init__(
        self,
        optimizer,
        init_lr,
        peak_lr,
        end_lr,
        warmup_steps=10000,
        total_steps=400000,
        current_step=0,
    ):
        self.init_lr = init_lr
        self.peak_lr = peak_lr
        self.end_lr = end_lr
        self.optimizer = optimizer
        self._warmup_rate = (peak_lr - init_lr) / warmup_steps
        self._decay_rate = (end_lr - peak_lr) / (total_steps - warmup_steps)
        self._current_step = current_step
        self.lr = init_lr
        self.warmup_steps = warmup_steps
        self.total_steps = total_steps
        self._last_lr = [self.lr]

    def set_lr(self, lr):
        self._last_lr = [g["lr"] for g in self.optimizer.param_groups]
        for g in self.optimizer.param_groups:
            # g['lr'] = lr
            g["lr"] = self.end_lr  ###锁定用线性

    def step(self):
        if self._current_step < self.warmup_steps:
            lr = self.init_lr + self._warmup_rate * self._current_step

        elif self._current_step > self.total_steps:
            lr = self.end_lr

        else:
            decay_ratio = (self._current_step - self.warmup_steps) / (
                self.total_steps - self.warmup_steps
            )
            if decay_ratio < 0.0 or decay_ratio > 1.0:
                raise RuntimeError(
                    "Decay ratio must be in [0.0, 1.0]. Fix LR scheduler settings."
                )
            coeff = 0.5 * (1.0 + math.cos(math.pi * decay_ratio))
            lr = self.end_lr + coeff * (self.peak_lr - self.end_lr)

        self.lr = lr = self.end_lr = 0.002  ###锁定用线性###不听话，直接锁定！
        self.set_lr(lr)
        self.lr = lr
        self._current_step += 1
        return self.lr


if __name__ == "__main__":
    m = nn.Linear(10, 10)
    opt = Adam(m.parameters(), lr=1e-4)
    s = WarmupCosineLRSchedule(
        opt, 1e-6, 2e-4, 1e-6, warmup_steps=2000, total_steps=20000, current_step=0
    )
    lrs = []
    for i in range(25000):
        s.step()
        lrs.append(s.lr)
        print(s.lr)

    plt.plot(lrs)
    plt.plot(range(0, 25000), lrs)
    plt.show()



import os
import traceback,gradio as gr
import logging
from tools.i18n.i18n import I18nAuto
from tools.my_utils import clean_path
i18n = I18nAuto()

logger = logging.getLogger(__name__)
import librosa,ffmpeg
import soundfile as sf
import torch
import sys
from mdxnet import MDXNetDereverb
from vr import AudioPre, AudioPreDeEcho
from bsroformer import BsRoformer_Loader

try:
    import gradio.analytics as analytics
    analytics.version_check = lambda:None
except:...

weight_uvr5_root = "tools/uvr5/uvr5_weights"
uvr5_names = []
for name in os.listdir(weight_uvr5_root):
    if name.endswith(".pth") or name.endswith(".ckpt") or "onnx" in name:
        uvr5_names.append(name.replace(".pth", "").replace(".ckpt", ""))

device=sys.argv[1]
is_half=eval(sys.argv[2])
webui_port_uvr5=int(sys.argv[3])
is_share=eval(sys.argv[4])

def html_left(text, label='p'):
    return f"""<div style="text-align: left; margin: 0; padding: 0;">
                <{label} style="margin: 0; padding: 0;">{text}</{label}>
                </div>"""

def html_center(text, label='p'):
    return f"""<div style="text-align: center; margin: 100; padding: 50;">
                <{label} style="margin: 0; padding: 0;">{text}</{label}>
                </div>"""

def uvr(model_name, inp_root, save_root_vocal, paths, save_root_ins, agg, format0):
    infos = []
    try:
        inp_root = clean_path(inp_root)
        save_root_vocal = clean_path(save_root_vocal)
        save_root_ins = clean_path(save_root_ins)
        is_hp3 = "HP3" in model_name
        if model_name == "onnx_dereverb_By_FoxJoy":
            pre_fun = MDXNetDereverb(15)
        elif model_name == "Bs_Roformer" or "bs_roformer" in model_name.lower():
            func = BsRoformer_Loader
            pre_fun = func(
                model_path = os.path.join(weight_uvr5_root, model_name + ".ckpt"),
                device = device,
                is_half=is_half
            )
        else:
            func = AudioPre if "DeEcho" not in model_name else AudioPreDeEcho
            pre_fun = func(
                agg=int(agg),
                model_path=os.path.join(weight_uvr5_root, model_name + ".pth"),
                device=device,
                is_half=is_half,
            )
        if inp_root != "":
            paths = [os.path.join(inp_root, name) for name in os.listdir(inp_root)]
        else:
            paths = [path.name for path in paths]
        for path in paths:
            inp_path = os.path.join(inp_root, path)
            if(os.path.isfile(inp_path)==False):continue
            need_reformat = 1
            done = 0
            try:
                info = ffmpeg.probe(inp_path, cmd="ffprobe")
                if (
                    info["streams"][0]["channels"] == 2
                    and info["streams"][0]["sample_rate"] == "44100"
                ):
                    need_reformat = 0
                    pre_fun._path_audio_(
                        inp_path, save_root_ins, save_root_vocal, format0,is_hp3
                    )
                    done = 1
            except:
                need_reformat = 1
                traceback.print_exc()
            if need_reformat == 1:
                tmp_path = "%s/%s.reformatted.wav" % (
                    os.path.join(os.environ["TEMP"]),
                    os.path.basename(inp_path),
                )
                os.system(
                    f'ffmpeg -i "{inp_path}" -vn -acodec pcm_s16le -ac 2 -ar 44100 "{tmp_path}" -y'
                )
                inp_path = tmp_path
            try:
                if done == 0:
                    pre_fun._path_audio_(
                        inp_path, save_root_ins, save_root_vocal, format0,is_hp3
                    )
                infos.append("%s->Success" % (os.path.basename(inp_path)))
                yield "\n".join(infos)
            except:
                infos.append(
                    "%s->%s" % (os.path.basename(inp_path), traceback.format_exc())
                )
                yield "\n".join(infos)
    except:
        infos.append(traceback.format_exc())
        yield "\n".join(infos)
    finally:
        try:
            if model_name == "onnx_dereverb_By_FoxJoy":
                del pre_fun.pred.model
                del pre_fun.pred.model_
            else:
                del pre_fun.model
                del pre_fun
        except:
            traceback.print_exc()
        print("clean_empty_cache")
        if torch.cuda.is_available():
            torch.cuda.empty_cache()
    yield "\n".join(infos)

with gr.Blocks(title="UVR5 WebUI") as app:
    gr.Markdown(
        value=
            i18n("本软件以MIT协议开源, 作者不对软件具备任何控制力, 使用软件者、传播软件导出的声音者自负全责. <br>如不认可该条款, 则不能使用或引用软件包内任何代码和文件. 详见根目录<b>LICENSE</b>.")
    )
    with gr.Group():
        gr.Markdown(html_center(i18n("伴奏人声分离&去混响&去回声"),'h2'))
        with gr.Group():
                gr.Markdown(
                    value=html_left(i18n("人声伴奏分离批量处理， 使用UVR5模型。") + "<br>" + \
                        i18n("合格的文件夹路径格式举例： E:\\codes\\py39\\vits_vc_gpu\\白鹭霜华测试样例(去文件管理器地址栏拷就行了)。")+ "<br>" + \
                        i18n("模型分为三类：") + "<br>" + \
                        i18n("1、保留人声：不带和声的音频选这个，对主人声保留比HP5更好。内置HP2和HP3两个模型，HP3可能轻微漏伴奏但对主人声保留比HP2稍微好一丁点；") + "<br>" + \
                        i18n("2、仅保留主人声：带和声的音频选这个，对主人声可能有削弱。内置HP5一个模型；") + "<br>" + \
                        i18n("3、去混响、去延迟模型（by FoxJoy）：") + "<br>  " + \
                        i18n("(1)MDX-Net(onnx_dereverb):对于双通道混响是最好的选择，不能去除单通道混响；") + "<br>&emsp;" + \
                        i18n("(234)DeEcho:去除延迟效果。Aggressive比Normal去除得更彻底，DeReverb额外去除混响，可去除单声道混响，但是对高频重的板式混响去不干净。") + "<br>" + \
                        i18n("去混响/去延迟，附：") + "<br>" + \
                        i18n("1、DeEcho-DeReverb模型的耗时是另外2个DeEcho模型的接近2倍；") + "<br>" + \
                        i18n("2、MDX-Net-Dereverb模型挺慢的；") + "<br>" + \
                        i18n("3、个人推荐的最干净的配置是先MDX-Net再DeEcho-Aggressive。"),'h4')
                )
                with gr.Row():
                    with gr.Column():
                        model_choose = gr.Dropdown(label=i18n("模型"), choices=uvr5_names)
                        dir_wav_input = gr.Textbox(
                            label=i18n("输入待处理音频文件夹路径"),
                            placeholder="C:\\Users\\Desktop\\todo-songs",
                        )
                        wav_inputs = gr.File(
                            file_count="multiple", label=i18n("也可批量输入音频文件, 二选一, 优先读文件夹")
                        )
                    with gr.Column():
                        agg = gr.Slider(
                            minimum=0,
                            maximum=20,
                            step=1,
                            label=i18n("人声提取激进程度"),
                            value=10,
                            interactive=True,
                            visible=False,  # 先不开放调整
                        )
                        opt_vocal_root = gr.Textbox(
                            label=i18n("指定输出主人声文件夹"), value="output/uvr5_opt"
                        )
                        opt_ins_root = gr.Textbox(
                            label=i18n("指定输出非主人声文件夹"), value="output/uvr5_opt"
                        )
                        format0 = gr.Radio(
                            label=i18n("导出文件格式"),
                            choices=["wav", "flac", "mp3", "m4a"],
                            value="flac",
                            interactive=True,
                        )
                        with gr.Column():
                            with gr.Row():
                                but2 = gr.Button(i18n("转换"), variant="primary")
                            with gr.Row():
                                vc_output4 = gr.Textbox(label=i18n("输出信息"),lines=3)
                    but2.click(
                        uvr,
                        [
                            model_choose,
                            dir_wav_input,
                            opt_vocal_root,
                            wav_inputs,
                            opt_ins_root,
                            agg,
                            format0,
                        ],
                        [vc_output4],
                        api_name="uvr_convert",
                    )
app.queue().launch(#concurrency_count=511, max_size=1022
    server_name="0.0.0.0",
    inbrowser=True,
    share=is_share,
    server_port=webui_port_uvr5,
    quiet=True,
)
